#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library list view implementation
"""

from __future__ import print_function, division, absolute_import

import traceback

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs import qt
from tpDcc.libs.qt.widgets.library import consts, mixin


class LibraryListView(mixin.LibraryViewWidgetMixin, QListView):
    """
    Class that implemented library list view widget
    This class is used by LibraryViewer class
    """

    DEFAULT_DRAG_THRESHOLD = consts.LIST_DEFAULT_DRAG_THRESHOLD

    itemMoved = Signal(object)
    itemDropped = Signal(object)
    itemClicked = Signal(object)
    itemDoubleClicked = Signal(object)

    def __init__(self, *args):
        QListView.__init__(self, *args)
        mixin.LibraryViewWidgetMixin.__init__(self)

        self.setSpacing(5)
        self.setMouseTracking(True)
        self.setSelectionRectVisible(True)
        self.setViewMode(QListView.IconMode)
        self.setResizeMode(QListView.Adjust)
        self.setSelectionMode(QListView.ExtendedSelection)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragDrop)

        self._tree_widget = None
        self._rubber_band = None
        self._rubber_band_start_pos = None
        self._rubber_band_color = QColor(Qt.white)
        self._custom_sort_order = list()

        self._drag = None
        self._drag_start_pos = None
        self._drag_start_index = None
        self._drop_enabled = True

        self.clicked.connect(self._on_index_clicked)
        self.doubleClicked.connect(self._on_index_double_clicked)

    """
    ##########################################################################################
    OVERRIDES
    ##########################################################################################
    """

    def mousePressEvent(self, event):
        """
        Overrides base QListView mousePressEvent function
        :param event: QMouseEvent
        """

        item = self.item_at(event.pos())
        if not item:
            self.clearSelection()

        mixin.LibraryViewWidgetMixin.mousePressEvent(self, event)
        if event.isAccepted():
            QListView.mousePressEvent(self, event)
            # self.viewer().tree_widget().setItemSelected(item, True)

        self.endDrag()
        self._drag_start_pos = event.pos()

        is_left_button = self.mouse_press_button() == Qt.LeftButton
        is_item_draggable = item and item.drag_enabled()
        is_selection_empty = not self.selected_items()

        if is_left_button and (is_selection_empty or not is_item_draggable):
            self.rubber_band_start_event(event)

    def mouseMoveEvent(self, event):
        """
        Overrides base QListView mouseMoveEvent function
        :param event: QMouseEvent
        """

        if not self.is_dragging_items():
            is_left_button = self.mouse_press_button() == Qt.LeftButton
            if is_left_button and self.rubber_band().isHidden() and self.selected_items():
                self.startDrag(event)
            else:
                mixin.LibraryViewWidgetMixin.mouseMoveEvent(self, event)
                QListView.mouseMoveEvent(self, event)
            if is_left_button:
                self.rubber_band_move_event(event)

    def mouseReleaseEvent(self, event):
        """
        Override base QListView mouseReleaseEvent function
        :param event: QMouseEvent
        """

        item = self.item_at(event.pos())
        items = self.selected_items()
        mixin.LibraryViewWidgetMixin.mouseReleaseEvent(self, event)
        if item not in items:
            if event.button() != Qt.MidButton:
                QListView.mouseReleaseEvent(self, event)
        elif not items:
            QListView.mouseReleaseEvent(self, event)

        self.endDrag()
        self.rubber_band().hide()

    def startDrag(self, event):
        """
        Overrides bae QListView startDrag function
        :param event: QEvent
        """

        if not self.dragEnabled():
            return

        if self._drag_start_pos and hasattr(event, 'pos'):
            item = self.item_at(event.pos())
            if item and item.drag_enabled():
                self._drag_start_index = self.indexAt(event.pos())
                point = self._drag_start_pos - event.pos()
                dt = self.drag_threshold()
                if point.x() > dt or point.y() > dt or point.x() < -dt or point.y() < -dt:
                    items = self.selected_items()
                    mime_data = self.mime_data(items)
                    pixmap = self._drag_pixmap(item, items)
                    hotspot = QPoint(pixmap.width() * 0.5, pixmap.height() * 0.5)
                    self._drag = QDrag(self)
                    self._drag.setPixmap(pixmap)
                    self._drag.setHotSpot(hotspot)
                    self._drag.setMimeData(mime_data)
                    self._drag.start(Qt.MoveAction)

    def endDrag(self):
        """
        Function that ends current drag
        """

        self._drag_start_pos = None
        self._drag_start_index = None
        if self._drag:
            del self._drag
            self._drag = None

    def dragEnterEvent(self, event):
        """
        Overrides bae QListView dragEnterEvent function
        :param event: QDragEvent
        """

        mimedata = event.mimeData()
        if (mimedata.hasText() or mimedata.hasUrls()) and self.drop_enabled():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """
        Overrides bae QListView dragMoveEvent function
        :param event: QDragEvent
        """

        mimedata = event.mimeData()
        if (mimedata.hasText() or mimedata.hasUrls()) and self.drop_enabled():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Overrides bae QListView dropEvent function
        :param event: QDropEvent
        """

        item = self.item_at(event.pos())
        selected_items = self.selected_item()
        if selected_items and item:
            if self.tree_widget().is_sort_by_custom_order():
                self.move_items(selected_items, item)
            else:
                qt.logger.info('You can only re-order items when sorting by custom order')

        if item:
            item.drop_event(event)

        self.itemDropped.emit(event)

    """
    ##########################################################################################
    BASE
    ##########################################################################################
    """

    def scroll_to_item(self, item, pos=None):
        """
        Ensures that the item is visible
        :param item: LibraryItem
        :param pos: QPoint or None
        """

        index = self.index_from_item(item)
        pos = pos or QAbstractItemView.PositionAtCenter

        self.scrollTo(index, pos)

    """
    ##########################################################################################
    EVENTS
    ##########################################################################################
    """

    def validate_drag_event(self, event):
        """
        Validates the drag event
        :param event: QMouseEvent
        """

        return Qt.LeftButton == event.mouseButtons()

    def rubber_band_start_event(self, event):
        """
        Function called when the user presses an empty area
        :param event: QMouseEvent
        """

        self._rubber_band_start_pos = event.pos()
        rect = QRect(self._rubber_band_start_pos, QSize())
        rubber_band = self.rubber_band()
        rubber_band.setGeometry(rect)
        rubber_band.show()

    def rubber_band_move_event(self, event):
        """
        Function called when the user moves the mouse over the current viewport
        :param event: QMouseEvent
        """

        if self.rubber_band() and self._rubber_band_start_pos:
            rect = QRect(self._rubber_band_start_pos, event.pos())
            rect = rect.normalized()
            self.rubber_band().setGeometry(rect)

    """
    ##########################################################################################
    DRAG & DROP
    ##########################################################################################
    """

    def drop_enabled(self):
        """
        Returns whether drop functionality is enabled or not
        :return: bool
        """

        return self._drop_enabled

    def set_drop_enabled(self, flag):
        """
        Sets whether drop functionality is enabled or not
        :param flag: bool
        """

        self._drop_enabled = flag

    def drag_threshold(self):
        """
        Returns current drag threshold
        :return: float
        """

        return self.DEFAULT_DRAG_THRESHOLD

    def is_dragging_items(self):
        """
        Returns whether the user is currently dragging items or not
        :return: bool
        """

        return bool(self._drag)

    def mime_data(self, items):
        """
        Returns drag mime data
        :param items: list(LibraryItem)
        :return: QMimeData
        """

        mimedata = QMimeData()
        urls = [item.url() for item in items]
        text = '\n'.join([item.mime_text() for item in items])
        mimedata.setUrls(urls)
        mimedata.setText(text)

        return mimedata

    def _drag_pixmap(self, item, items):
        """
        Internal function that shows the pixmap for the given item during drag operation
        :param item: LibraryItem
        :param items: list(LibraryItem)
        :return: QPixmap
        """

        rect = self.visualRect(self.index_from_item(item))
        pixmap = QPixmap()
        pixmap = pixmap.grabWidget(self, rect)
        if len(items) > 1:
            custom_width = 35
            custom_padding = 5
            custom_text = str(len(items))
            custom_x = pixmap.rect().center().x() - float(custom_width * 0.5)
            custom_y = pixmap.rect().top() + custom_padding
            custom_rect = QRect(custom_x, custom_y, custom_width, custom_width)
            painter = QPainter(pixmap)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.setPen(Qt.NoPen)
            painter.setBrush(self.viewer().background_selected_color())
            painter.drawEllipse(custom_rect.center(), float(custom_width * 0.5), float(custom_width * 0.5))
            font = QFont('Serif', 12, QFont.Light)
            painter.setFont(font)
            painter.setPen(self.viewer().text_selected_color())
            painter.drawText(custom_rect, Qt.AlignCenter, str(custom_text))

        return pixmap

    """
    ##########################################################################################
    TREE WIDGET
    ##########################################################################################
    """

    def tree_widget(self):
        """
        Return the tree widget that contains the items
        :return: LibraryTreeWidget
        """

        return self._tree_widget

    def set_tree_widget(self, tree_widget):
        """
        Set the tree widget that contains the items
        :param tree_widget: LibraryTreeWidget
        """

        self._tree_widget = tree_widget
        self.setModel(tree_widget.model())
        self.setSelectionModel(tree_widget.selectionModel())

    def items(self):
        """
        Return all the items
        :return: list(LibraryItem)
        """

        return self.tree_widget().items()

    def item_at(self, pos):
        """
        Returns a pointer to the item at the coordinates p
        The coordinates are relative to the tree widget's viewport
        :param pos: QPoint
        :return: LibraryItem
        """

        index = self.indexAt(pos)
        return self.item_from_index(index)

    def selected_item(self):
        """
        Returns the last selected non-hidden item
        :return: QTreeWidgetItem
        """

        return self.tree_widget().selected_item()

    def selected_items(self):
        """
        Returns a list of all selected non-hidden items
        :return: list(QTreeWidgetItem)
        """

        return self.tree_widget().selectedItems()

    def insert_item(self, row, item):
        """
        Inserts the item at row in the top level in the view
        :param row: int
        :param item: QTreeWidgetItem
        """

        self.tree_widget().insertTopLevelItem(row, item)

    def take_items(self, items):
        """
        Removes and returns the items from the view
        :param items: list(QTreeWidgetItem)
        :return: list(QTreeWidgetItem)
        """

        for item in items:
            row = self.tree_widget().indexOfTopLevelItem(item)
            self.tree_widget().takeTopLevelItem(row)

        return items

    def set_indexes_selected(self, indexes, value):
        """
        Set the selected state for the given indexes
        :param indexes: list(QModelIndex)
        :param value: bool
        """

        items = self.items_from_indexes(indexes)
        self.set_items_selected(items, value)

    def set_items_selected(self, items, value):
        """
        Sets the selected state for the given items
        :param items: list(LibraryItem)
        :param value: bool
        """

        self.tree_widget().blockSignals(True)
        try:
            for item in items:
                self.tree_widget().setItemSelected(item, value)
        except Exception as e:
            qt.logger.error('{} | {}'.format(e, traceback.format_exc()))
        finally:
            self.tree_widget().blockSignals(False)

    def move_items(self, items, item_at):
        """
        Moves the given items to the position at the given row
        :param items: list(LibraryItem)
        :param item_at: LibraryItem
        """

        scroll_value = self.verticalScrollBar().value()
        self.tree_widget().move_items(items, item_at)
        self.itemMoved.emit(items[-1])
        self.verticalScrollBar().setValue(scroll_value)

    def index_from_item(self, item):
        """
        Returns QModelIndex associated with the given item
        :param item: LibraryItem
        :return: QModelIndex
        """

        return self.tree_widget().indexFromItem(item)

    def item_from_index(self, index):
        """
        Return a pointer to the LibraryItem associated with the given model index
        :param index: QModelIndex
        :return: LibraryItem
        """

        return self.tree_widget().itemFromIndex(index)

    """
    ##########################################################################################
    RUBBER BAND
    ##########################################################################################
    """

    def create_rubber_band(self):
        """
        Creates a new instance of the selection rubber band
        :return: QRubberBand
        """

        rubber_band = QRubberBand(QRubberBand.Rectangle, self)
        palette = QPalette()
        color = self.rubber_band_color()
        palette.setBrush(QPalette.Highlight, QBrush(color))
        rubber_band.setPalette(palette)

        return rubber_band

    def rubber_band(self):
        """
        Retursn the selection rubber band for this widget
        :return: QRubberBand
        """

        if not self._rubber_band:
            self.setSelectionRectVisible(False)
            self._rubber_band = self.create_rubber_band()

        return self._rubber_band

    def rubber_band_color(self):
        """
        Returns the rubber band color for this widget
        :return: QColor
        """

        return self._rubber_band_color

    def set_rubber_band_color(self, color):
        """
        Sets the color for the rubber band
        :param color: QColor
        """

        self._rubber_band = None
        self._rubber_band_color = color

    """
    ##########################################################################################
    CALLBACKS
    ##########################################################################################
    """

    def _on_index_clicked(self, index):
        """
        Callback function that is called when the user clicks on an item
        :param index: QModelIndex
        """

        item = self.item_from_index(index)
        item.clicked()
        self.set_items_selected([item], True)
        self.itemClicked.emit(item)

    def _on_index_double_clicked(self, index):
        """
        Callback function that is called when the user double clicks on an item
        :param index: QModelIndex
        """

        item = self.item_from_index(index)
        self.set_items_selected([item], True)
        item.double_clicked()
        self.itemDoubleClicked.emit(item)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt spinner widgets
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.qt.core import base, mixin, theme
from tpDcc.libs.qt.widgets import lineedit, buttons, label


@mixin.theme_mixin
@mixin.cursor_mixin
class BaseSpinBox(QSpinBox, object):
    def __init__(self, parent=None):
        super(BaseSpinBox, self).__init__(parent=parent)

        self._size = self.theme_default_size()

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_size(self):
        """
        Returns the spin box height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets spin box height size
        :param value: float
        """

        self._size = value
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def tiny(self):
        """
        Sets spin box to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets spin box to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets spin box to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets spin box to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets spin box to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self


@mixin.theme_mixin
@mixin.cursor_mixin
class BaseDoubleSpinBox(QDoubleSpinBox, object):
    def __init__(self, parent=None):
        super(BaseDoubleSpinBox, self).__init__(parent=parent)

        self._size = self.theme_default_size()

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_size(self):
        """
        Returns the double spin box height size
        :return: float
        """

        return self._size

    def _set_size(self, value):
        """
        Sets double spin box height size
        :param value: float
        """

        self._size = value
        self.style().polish(self)

    theme_size = Property(int, _get_size, _set_size)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def tiny(self):
        """
        Sets double spin box to tiny size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.tiny if widget_theme else theme.Theme.Sizes.TINY

        return self

    def small(self):
        """
        Sets double spin box to small size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.small if widget_theme else theme.Theme.Sizes.SMALL

        return self

    def medium(self):
        """
        Sets double spin box to medium size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.medium if widget_theme else theme.Theme.Sizes.MEDIUM

        return self

    def large(self):
        """
        Sets double spin box to large size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.large if widget_theme else theme.Theme.Sizes.LARGE

        return self

    def huge(self):
        """
        Sets double spin box to huge size
        """

        widget_theme = self.theme()
        self.theme_size = widget_theme.huge if widget_theme else theme.Theme.Sizes.HUGE

        return self


class BaseNumberWidget(base.BaseWidget, object):
    valueChanged = Signal(object)

    def __init__(self, name='', parent=None):
        self._name = name
        super(BaseNumberWidget, self).__init__(parent)

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        return main_layout

    def ui(self):
        super(BaseNumberWidget, self).ui()

        self._number_widget = self.get_number_widget()
        self._number_label = label.BaseLabel(self._name, parent=self)
        self._number_label.setAlignment(Qt.AlignLeft)
        self._number_label.setMinimumWidth(75)
        self._number_label.setAttribute(Qt.WA_TransparentForMouseEvents)
        if not self._name:
            self._number_label.hide()
        self._value_label = label.BaseLabel('value', parent=self)
        self._value_label.setAlignment(Qt.AlignLeft)
        self._value_label.setMinimumWidth(75)
        self._value_label.setAttribute(Qt.WA_TransparentForMouseEvents)
        self._value_label.hide()

        self.main_layout.addWidget(self._number_label)
        self.main_layout.addSpacing(5)
        self.main_layout.addWidget(self._value_label, alignment=Qt.AlignRight)
        self.main_layout.addWidget(self._number_widget)

    def get_number_widget(self):
        """
        Returns the widget used to edit numeric value
        :return: QWidget
        """

        spin_box = BaseSpinBox(parent=self)
        spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        return spin_box

    def get_value(self):
        """
        Returns the number value of the numeric widget
        :return: variant, int || float
        """

        return self._number_widget.value()

    def set_value(self, new_value):
        """
        Sets the value of the numeric widget
        :param new_value: variant, int || float
        """

        if new_value:
            self._number_widget.setValue(new_value)

    def get_label_text(self):
        return self._number_label.text()

    def set_label_text(self, new_text):
        self._number_label.setText(new_text)

    def set_value_label(self, new_value):
        self._value_label.show()
        self._value_label.setText(str(new_value))

    def _on_value_changed(self):
        self.valueChanged.emit(self.get_value())


class BaseSpinBoxNumber(BaseNumberWidget):
    enterPressed = Signal()

    def __init__(self, name='', parent=None):
        super(BaseSpinBoxNumber, self).__init__(name=name, parent=parent)

        self._setup_spin_widget()

    def keyPressEvent(self, event):
        if event.key() in [Qt.Key_Return, Qt.Key_Enter]:
            self.enterPressed.emit()

    def get_number_widget(self):
        spin_box = BaseSpinBox()
        spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        return spin_box

    def _setup_spin_widget(self):
        if hasattr(self._number_widget, 'CorrectToNearestValue'):
            self._number_widget.setCorrectionMode(self._number_widget.CorrectToNearestValue)
        if hasattr(self._number_widget, 'setWrapping'):
            self._number_widget.setWrapping(False)
        if hasattr(self._number_widget, 'setDecimals'):
            self._number_widget.setDecimals(3)

        self._number_widget.setMaximum(100000000)
        self._number_widget.setMinimum(-100000000)
        # self._number_widget.setButtonSymbols(self._number_widget.NoButtons)

        self._number_widget.valueChanged.connect(self._on_value_changed)


class BaseDoubleNumberSpinBox(BaseSpinBoxNumber, object):
    def __init__(self, name='', parent=None):
        super(BaseDoubleNumberSpinBox, self).__init__(name=name, parent=parent)

    def get_number_widget(self):
        spin_box = BaseDoubleSpinBox()
        spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        return spin_box


class DragDoubleSpinBox(BaseNumberWidget, object):
    def __init__(self, name='', parent=None):
        super(DragDoubleSpinBox, self).__init__(name=name, parent=parent)

    def get_number_widget(self):
        spin_box = DragDoubleSpinBoxLine()
        spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        return spin_box


class DragDoubleSpinBoxLine(lineedit.BaseLineEdit, object):
    """
    Using middle mouse from left to right will scale the value and a little bar will show the
    percent of the current value
    """

    valueChanged = Signal(float)

    def __init__(self, start=0.0, max=10, min=-10, positive=False, parent=None):
        super(DragDoubleSpinBoxLine, self).__init__(parent=parent)

        self._click = False
        self._default = str(start)
        self._mouse_position = QPoint(0, 0)
        self._min = 0.01 if positive else min
        self._max = max
        self._sup = positive
        self.setText(str(start))

        self._setup_validator()

        theme = self.theme()
        self._color = theme.accent_color_5 or QColor(0, 255, 0)

    def mousePressEvent(self, event):
        if event.button() == Qt.MiddleButton:
            self._click = True
            self._mouse_position = event.pos()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.MiddleButton:
            self._click = False

    def mouseDoubleClickEvent(self, event):
        self.setText(self._default)

    def mouseMoveEvent(self, event):
        if self._click:
            delta = event.x() - self._mouse_position.x()
            v = float(self.text()) + delta / 100.0
            v = max(self._min, min(self._max, v))
            self.setText(str(v))
            self._mouse_position = event.pos()
            self.valueChanged.emit(v)

    def paintEvent(self, event):
        super(DragDoubleSpinBoxLine, self).paintEvent(event)
        p = QPainter()
        p.begin(self)

        try:
            v = float(self.text())
        except Exception:
            v = 0.0000001

        try:
            v /= self._max if v > 0 else (self._min * -1)
        except Exception:
            pass
        if self._sup:
            p.fillRect(QRect(0, self.height() - 4, v * self.width(), 4), self._color)
        else:
            p.fillRect(
                QRect(self.width() * 0.5, self.height() - 4, v * self.width() * 0.5, 4),
                self._color if v > 0 else QColor(255, 0, 0))
        p.end()

    def get_validator(self):
        return QDoubleValidator()

    def value(self):
        try:
            return float(self.text())
        except Exception:
            return 0.0

    # NOTE: Here I'm breaking naming nomenclature on purpose. Doing this we follow nomenclature of Qt SpinBoxes
    def setValue(self, new_value):
        self.setText(str(new_value))

    def _setup_validator(self):
        self.setValidator(self.get_validator())


class DragDoubleSpinBoxLineAxis(base.BaseWidget, object):

    textChanged = Signal(str)
    valueChanged = Signal(float)

    def __init__(self, axis, reset=True, start=0.0, max=10, min=-10, positive=False, parent=None):
        self._axis = axis
        self._reset = reset
        self._start = start
        self._max = max
        self._min = min
        self._positive = positive
        super(DragDoubleSpinBoxLineAxis, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        return main_layout

    def ui(self):
        super(DragDoubleSpinBoxLineAxis, self).ui()

        axis_widget = QFrame()
        axis_widget.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        axis_layout = QHBoxLayout()
        axis_layout.setContentsMargins(0, 0, 0, 0)
        axis_layout.setSpacing(0)
        self._axis_btn = buttons.get_axis_button(axis_type=self._axis, parent=self)
        self._line = DragDoubleSpinBoxLine(
            start=self._start, max=self._max, min=self._min, positive=self._positive, parent=self)
        self._reset_btn = buttons.BaseToolButton().image('reset').icon_only()
        self._reset_btn.clicked.connect(partial(self._line.setValue, self._start))
        self._reset_btn.setVisible(self._reset)
        self._reset_btn.setEnabled(self._reset)
        axis_layout.addWidget(self._axis_btn)
        axis_layout.addWidget(self._line)
        axis_layout.addWidget(self._reset_btn)
        axis_widget.setLayout(axis_layout)

        self.main_layout.addWidget(axis_widget)

    def setup_signals(self):
        self._line.valueChanged.connect(self.valueChanged.emit)
        self._line.textChanged.connect(self.textChanged.emit)

    def value(self):
        return self._line.value()

    def setValue(self, new_value):
        self._line.setValue(new_value)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt tree widgets
"""

from __future__ import print_function, division, absolute_import

import string

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs import qt
from tpDcc.libs.python import path, fileio, folder
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import buttons, search


class TreeWidget(QTreeWidget, object):

    ITEM_WIDGET = QTreeWidgetItem
    ITEM_WIDGET_SIZE = None

    def __init__(self, parent=None):
        super(TreeWidget, self).__init__(parent)

        self._auto_add_sub_items = True
        self._title_text_index = 0
        self._text_edit = True
        self._edit_state = None
        self._current_name = None
        self._old_name = None
        self._current_item = None
        self._last_item = None
        self._drop_indicator_rect = QRect()

        self.setIndentation(25)
        self.setExpandsOnDoubleClick(False)

        if tp.Dcc.get_name() == tp.Dccs.Maya:
            self.setAlternatingRowColors(tp.Dcc.get_version() < 2016)

        self.setSortingEnabled(True)
        self.sortByColumn(0, Qt.AscendingOrder)
        if not tp.is_maya() and not not tp.is_nuke():
            palette = QPalette()
            palette.setColor(palette.Highlight, Qt.gray)
            self.setPalette(palette)

        self.itemCollapsed.connect(self._on_item_collapsed)
        self.itemActivated.connect(self._on_item_activated)
        self.itemChanged.connect(self._on_item_changed)
        self.itemSelectionChanged.connect(self._on_item_selection_changed)
        self.itemClicked.connect(self._on_item_clicked)
        self.itemExpanded.connect(self._on_item_expanded)

    @property
    def current_item(self):
        return self._current_item

    @property
    def current_name(self):
        return self._current_name

    @property
    def edit_state(self):
        return self._edit_state

    @edit_state.setter
    def edit_state(self, flag):
        self._edit_state = flag

    def paintEvent(self, event):
        painter = QPainter(self.viewport())
        self.drawTree(painter, event.region())
        self._paint_drop_indicator(painter)

    def mousePressEvent(self, event):

        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.AltModifier:
            pos = self.mapToGlobal((self.rect().topLeft()))
            QWhatsThis.showText(pos, self.whatsThis())
            return

        super(TreeWidget, self).mousePressEvent(event)

        item = self.itemAt(event.pos())
        if not item:
            self._clear_selection()
        else:
            self._current_item = item

    def mouseDoubleClickEvent(self, event):
        position = event.pos()
        index = self.indexAt(position)
        self.doubleClicked.emit(index)

    def dragMoveEvent(self, event):
        item = self.itemAt(event.pos())

        if item:
            index = self.indexFromItem(item)
            rect = self.visualRect(index)
            rect_left = self.visualRect(index.sibling(index.row(), 0))
            rect_right = self.visualRect(index.sibling(index.row(), self.header().logicalIndex(self.columnCount() - 1)))
            self._drop_indicator_position = self.position(event.pos(), rect, index)
            if self._drop_indicator_position == self.AboveItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.top(), rect_right.right() - rect_left.left(), 0)
                event.accept()
            elif self._drop_indicator_position == self.BelowItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.bottom(), rect_right.right() - rect_left.left(), 0)
                event.accept()
            elif self._drop_indicator_position == self.OnItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.top(), rect_right.right() - rect_left.left(), rect.height())
                event.accept()
            else:
                self._drop_indicator_rect = QRect()

            self.model().setData(index, self._drop_indicator_position, Qt.UserRole)

        self.viewport().update()

        super(TreeWidget, self).dragMoveEvent(event)

    def addTopLevelItem(self, item):
        super(TreeWidget, self).addTopLevelItem(item)

        if hasattr(item, 'widget'):
            if hasattr(item, 'column'):
                self.setItemWidget(item, item.column, item.widget)
            else:
                self.setItemWidget(item, 0, item.widget)

    def insertTopLevelItem(self, index, item):
        super(TreeWidget, self).insertTopLevelItem(index, item)

        if hasattr(item, 'widget'):
            if hasattr(item, 'column'):
                self.setItemWidget(item, item.column, item.widget)
            else:
                self.setItemWidget(item, 0, item.widget)

    def position(self, pos, rect, index):
        """
        Function that returns whether the cursor is over, below or on an item
        :param pos: QPos
        :param rect: QRect
        :param rect: QModelIndex
        :return: QAbstractItemView.DropIndicatorPosition
        """

        r = QAbstractItemView.OnViewport

        # NOTE: margin * 2 MUST be smaller than row height, otherwise drop OnItem rect will not show
        margin = 5

        if pos.y() - rect.top() < margin:
            r = QAbstractItemView.AboveItem
        elif rect.bottom() - pos.y() < margin:
            r = QAbstractItemView.BelowItem
        elif pos.y() - rect.top() > margin and rect.bottom() - pos.y() > margin:
            r = QAbstractItemView.OnItem

        return r

    def unhide_items(self):
        """
        Unhide all tree items
        """

        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            self.setItemHidden(item, False)

    def filter_names(self, filter_text):
        """
        Hides all tree items with the given text
        :param filter_text: str, text used to filter tree items
        """

        self.unhide_items()

        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            text = str(item.text(self._title_text_index))
            filter_text = str(filter_text)

            # If the filter text is not found on the item text, we hide the item
            if text.find(filter_text) == -1:
                self.setItemHidden(item, True)

    def get_tree_item_name(self, tree_item):
        """
        Returns a list with all the column names of the given QTreeWidgetItem
        :param tree_item: QTreeWidgetItem, item we want to retrive name of
        :return: list<str>
        """

        try:
            # When selecting an item in the tree and refreshing a C++ wrapped will raise
            count = QTreeWidgetItem.columnCount(tree_item)
        except Exception:
            count = 0

        name = list()
        for i in range(count):
            name.append(str(tree_item.text(i)))

        return name

    def get_tree_item_path_string(self, tree_item):
        """
        Returns full path of the given QTreeWidgetItem from its parent into it with the following format:
        "parent/parent/item"
        :param tree_item:
        :return:
        """

        parents = self.get_tree_item_path(tree_item)
        parent_names = self.get_tree_item_names(parents)

        if not parent_names:
            return
        if len(parent_names) == 1 and not parent_names[0]:
            return

        names = list()

        for name in parent_names:
            names.append(name[0])

        names.reverse()

        item_path_str = string.join(names, '/')

        return item_path_str

    def get_tree_item_path(self, tree_item):
        """
        Return the tree path of the given QTreeWidgetItem in a list starting from the given item
        :param tree_item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        if not tree_item:
            return

        parent_items = list()
        parent_items.append(tree_item)

        try:
            # When selecting an item in the tree and refreshing a C++ wrapped will raise
            parent_item = tree_item.parent()
        except Exception:
            parent_item = None

        while parent_item:
            parent_items.append(parent_item)
            parent_item = parent_item.parent()

        return parent_items

    def get_tree_item_names(self, tree_items):
        """
        Returns a list with the names of the given QTreeWidgetItems
        :param tree_items: list<QTreeWidgetItem>
        :return: list<str>
        """

        item_names = list()
        if not tree_items:
            return item_names

        for tree_item in tree_items:
            name = self.get_tree_item_name(tree_item)
            if name:
                item_names.append(name)

        return item_names

    def get_tree_item_children(self, tree_item):
        """
        Returns all child items of the given QTreeWidgetItem
        :param tree_item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        count = tree_item.childCount()
        items = list()
        for i in range(count):
            items.append(tree_item.child(i))

        return items

    def delete_empty_children(self, tree_item):
        """
        Deletes all given QTreeWidget child items that are empty (has no text)
        :param tree_item: QTreeWidgetItem
        """

        count = tree_item.childCount()
        if count <= 0:
            return

        for i in range(count):
            item = tree_item.child(i)
            if item and not item.text(0):
                item = tree_item.takeChild(i)
                del item

    def delete_tree_item_children(self, tree_item):
        """
        Deletes all given QTreeWidget chlid items
        :param tree_item: QTreeWidgetItem
        """

        count = tree_item.childCount()
        if count <= 0:
            return

        children = tree_item.takeChildren()
        for child in children:
            del child

    def _drop_on(self, event_list):
        """
        Internal function that checks whether or not event list contains a valid drop operation
        :param event_list: list
        :return: bool
        """

        event, row, col, index = event_list
        root = self.rootIndex()

        if self.viewport().rect().contains(event.pos()):
            index = self.indexAt(event.pos())
            if not index.isValid() or not self.visualRect(index).contains(event.pos()):
                index = root

        if index != root:
            drop_indicator_position = self.position(event.pos(), self.visualRect(index), index)
            if self._drop_indicator_position == self.AboveItem:
                # Drop Above item
                row = index.row()
                col = index.column()
                index = index.parent()
            elif self._drop_indicator_position == self.BelowItem:
                # Drop Below item
                row = index.row() + 1
                col = index.column()
                index = index.parent()
        else:
            self._drop_indicator_position = self.OnViewport

        # Update given referenced list
        event_list[0], event_list[1], event_list[2], event_list[3] = event, row, col, index

        return True

    def _is_item_dropped(self, event, strict=False):
        """
        Returns whether or not an item has been dropped in given event
        :param event: QDropEvent
        :param strict: bool, True to handle ordered alphabetically list; False otherwise.
        :return: bool
        """

        is_dropped = False

        pos = event.pos()
        index = self.indexAt(pos)

        if event.source == self and event.dropAction() == Qt.MoveAction or \
                self.dragDropMode() == QAbstractItemView.InternalMove:
            top_index = QModelIndex()
            col = -1
            row = -1
            event_list = [event, row, col, top_index]
            if self._drop_on(event_list):
                event, row, col, top_index = event_list
                if row > -1:
                    if row == index.row() - 1:
                        is_dropped = False
                if row == -1:
                    is_dropped = True
                if row == index.row() + 1:
                    is_dropped = False if strict else True

        return is_dropped

    def _paint_drop_indicator(self, painter):
        """
        Internal function used to paint the drop indicator manually
        :param painter: QPainter
        """

        if self.state() == QAbstractItemView.DraggingState:
            opt = QStyleOption()
            opt.initFrom(self)
            opt.rect = self._drop_indicator_rect
            rect = opt.rect

            color = Qt.black
            if tp.is_maya():
                color = Qt.white

            brush = QBrush(QColor(color))
            pen = QPen(brush, 1, Qt.DotLine)
            painter.setPen(pen)
            if rect.height() == 0:
                painter.drawLine(rect.topLeft(), rect.topRight())
            else:
                painter.drawRect(rect)

    def _edit_start(self, item):
        """
        Internal function that is called when a user start editing a tree item text
        Closes already opened edit text editors and updates internal variables
        :param item: QTreeWidgetItem
        """

        self._old_name = str(item.text(self._title_text_index))
        self.closePersistentEditor(item, self._title_text_index)
        self.openPersistentEditor(item, self._title_text_index)
        self._edit_state = item

        return

    def _edit_finish(self, item):
        """
        Internal function that is called when a text element of the tree is edited
        Checks that that edit mode is on text and updates the item text manually
        :param item: QTreeWidgetItem
        :return: QTreeWidgetItem, edited item
        """

        if not hasattr(self._edit_state, 'text'):
            return

        self._edit_state = None

        if type(item) == int:
            return self._current_item

        self.closePersistentEditor(item, self._title_text_index)
        state = self._item_rename_valid(self._old_name, item)

        if state:
            state = self._item_renamed(item)
            if not state:
                item.setText(self._title_text_index, self._old_name)
                return item
        else:
            item.setText(self._title_text_index, self._old_name)
            return item

        return item

    def _item_rename_valid(self, old_name, item):
        """
        Checks if the rename operation on a specific item of the tree is valid or not
        :param old_name: str, old name of the item
        :param item: QTreeWidgetItem, item that is being edit
        :return: bool
        """

        new_name = item.text(self._title_text_index)
        if not new_name:
            return False

        # We do not allow duplicated names on the tree
        if self._already_exists(item):
            return False

        if old_name == new_name:
            return False

        return True

    def _already_exists(self, item):
        """
        Checks if a given QTreeWidgetItem already exists on the tree.
        :param item: QTreeWidgetItem
        :return: bool
        """

        name = item.text(0)
        parent = item.parent()

        if parent:
            skip_index = parent.indexOfChild(item)
            for i in range(parent.childCount()):
                if i == skip_index:
                    continue
                other_name = str(parent.child(i).text(0))
                if name == other_name:
                    return True
        else:
            skip_index = self.indexFromItem(item)
            skip_index = skip_index.row()
            for i in range(self.topLevelItemCount()):
                if skip_index == i:
                    continue
                other_name = str(self.topLevelItem(i).text(0))
                if name == other_name:
                    return True

        return False

    def _clear_selection(self):
        """
        Internal function used to clear the selection of the tree and update internal variables
        """

        self.clearSelection()
        self._current_item = None
        if self._edit_state:
            self._edit_finish(self._last_item)

    def _emit_item_click(self, item):
        """
        Internal function that is used to force the emission of itemClicked signal
        :param item: QTreeWidgetItem
        """

        if item:
            name = item.text(self._title_text_index)
        else:
            name = ''

        self.itemClicked.emit(item, 0)

    def _get_ancestors(self, item):
        """
        Returns all ancestors items of the given QTreeWidgetItem
        :param item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        child_count = item.childCount()
        items = list()
        for i in range(child_count):
            child = item.child(i)
            children = self._get_ancestors(child)
            items.append(child)
            if children:
                items += children

        return items

    def _add_sub_items(self, tree_item):
        """
        Internal function that is updates the hiearchy of the given QTreeWidgetItem
        Implementation MUST be implemented in child class
        :param tree_item: QTreeWidgetItem
        """

        pass

    def _item_renamed(self, item):
        """
        Internal function that rename a specific QTreeWidgetItem contents
        Implementation MUST be implemented in child class
        :param item: QTreeWidgetItem
        """

        return False

    def _on_item_expanded(self, item):
        """
        Internal function that is called anytime the user expands an item of the tree
        Load dynamically all the items parented to the expanded item
        :param item: QTreeWidgetItem
        """

        if self._auto_add_sub_items:
            self._add_sub_items(item)

    def _on_item_collapsed(self, item):
        """
        Internal function that is called anytime the user collapses an item of the tree
        :param item: QTreeWidgetItem
        """

        pass

    def _on_item_activated(self, item):
        """
        Internal function that is called anytime a tree item is activated
        :param item: QTreeWidgetItem
        """

        if self._edit_state:
            self._edit_finish(self._edit_state)
        else:
            if self._text_edit:
                self._edit_start(item)

    def _on_item_changed(self, current_item, previous_item):
        """
        Internal function that is called when a tree item changes its content
        :param current_item: QTreeWidetItem
        :param previous_item: QTreeWidgetItem
        """

        if self._edit_state:
            self._edit_finish(previous_item)

    def _on_item_selection_changed(self):
        """
        Internal function that is called anytime the user changes the selection of the tree
        """

        item_sel = self.selectedItems()
        current_item = None
        if item_sel:
            current_item = item_sel[0]

        if current_item:
            self._current_name = current_item.text(self._title_text_index)
        if self._edit_state:
            self._edit_finish(self._edit_state)

        if not current_item:
            self._emit_item_click(current_item)

    def _on_item_clicked(self, item, column):
        """
        Internal function that is called anytime the user clicks on an item of the tree
        Updates internal variables and clear the selection if necessary
        :param item: QTreeWidgetItem
        :param column: int
        """

        self._last_item = self._current_item
        self._current_item = self.currentItem()
        if not item or column != self._title_text_index:
            if self._last_item:
                self._clear_selection()


class ManageTreeWidget(base.BaseWidget, object):
    def __init__(self, parent=None):
        self.tree_widget = None
        super(ManageTreeWidget, self).__init__(parent)

    def set_tree_widget(self, tree_widget):
        """
        Set the tree widget managed by this widget
        :param tree_widget: QTreeWidget
        """

        self.tree_widget = tree_widget


class FilterTreeWidget(base.DirectoryWidget, object):
    NameFilterChanged = Signal(object)

    def __init__(self, parent=None):
        self.tree_widget = None
        self.emit_changes = True
        self.update_tree = True
        super(FilterTreeWidget, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        return main_layout

    def ui(self):
        super(FilterTreeWidget, self).ui()

        self.filter_names = search.SearchFindWidget()
        self.filter_names.set_placeholder_text('Filter Names')
        self.main_layout.addWidget(self.filter_names)

    def setup_signals(self):
        self.filter_names.textChanged.connect(self._on_filter_names)

    def get_name_filter(self):
        """
        Returns the name filter current text
        :return: str
        """

        return str(self.filter_names.text())

    def set_emit_changes(self, flag):
        """
        Sets whether signals should be emitted or not
        :param flag: bool
        """

        self.emit_changes = flag

    def set_name_filter(self, text):
        """
        Sets the name filter text
        :param text: str
        """

        self.filter_names.setText(text)

    def clear_name_filter(self):
        """
        Clears current name filter text
        """

        self.filter_names.setText('')

    def set_tree_widget(self, tree_widget):
        """
        Sets the tree widget used by this widget
        :param tree_widget: QTreeWidget
        """

        self.tree_widget = tree_widget

    def _on_filter_names(self, text):
        """
        Internal function that is used to call the filter function of the TreeWidget
        :param text: str, filter text that should be used
        """

        if self.update_tree:
            self.tree_widget.filter_names(text)


class FilterTreeDirectoryWidget(base.DirectoryWidget, object):
    SubPathChanged = Signal(object)
    NameFilterChanged = Signal(object)

    def __init__(self, parent=None):
        self.tree_widget = None
        self.emit_changes = True
        self.update_tree = True
        super(FilterTreeDirectoryWidget, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        return main_layout

    def ui(self):
        super(FilterTreeDirectoryWidget, self).ui()

        self.filter_names = search.SearchFindWidget()
        self.filter_names.set_placeholder_text('Filter Names')
        # self.filter_names.setPlaceholderText('Filter Names')

        self.sub_path_filter = QLineEdit()
        self.sub_path_filter.setPlaceholderText('Set Sub Path')
        self.sub_path_filter.setVisible(False)

        self.main_layout.addWidget(self.filter_names)
        self.main_layout.addWidget(self.sub_path_filter)

    def setup_signals(self):
        self.sub_path_filter.textChanged.connect(self._on_sub_path_filter_changed)
        self.sub_path_filter.textEdited.connect(self._on_sub_path_filter_edited)
        self.filter_names.textChanged.connect(self._on_filter_names)

    def get_name_filter(self):
        """
        Returns the name filter current text
        :return: str
        """

        return str(self.filter_names.text())

    def get_sub_path_filter(self):
        """
        Returns the sub path filter current text
        :return: str
        """

        return str(self.sub_path_filter.text())

    def set_emit_changes(self, flag):
        """
        Sets whether signals should be emitted or not
        :param flag: bool
        """

        self.emit_changes = flag

    def set_name_filter(self, text):
        """
        Sets the name filter text
        :param text: str
        """

        self.filter_names.setText(text)

    def set_sub_path_filter(self, text):
        """
        Sets the sub path filter text
        :param text: str
        """

        self.sub_path_filter.setText(text)

    def clear_name_filter(self):
        """
        Clears current name filter text
        """

        self.filter_names.setText('')

    def clear_sub_path_filter(self):
        """
        Clears current sub path filter text
        """

        self.sub_path_filter.setText('')

    def set_tree_widget(self, tree_widget):
        """
        Sets the tree widget used by this widget
        :param tree_widget: QTreeWidget
        """

        self.tree_widget = tree_widget

    def set_sub_path_warning(self, flag):
        """
        Updates the style of the sub path QLineEdit depending on the app that uses this widget
        :param flag: bool
        """

        if flag:
            if not tp.is_maya():
                self.sub_path_filter.setStyleSheet('background-color: rgb(255, 150, 150);')
            else:
                self.sub_path_filter.setStyleSheet('background-color: rgb(255, 100, 100);')
        else:
            self.sub_path_filter.setStyleSheet('')

    def _on_filter_names(self, text):
        """
        Internal function that is used to call the filter function of the TreeWidget
        :param text: str, filter text that should be used
        """

        if self.update_tree:
            self.tree_widget.filter_names(text)

    def _on_sub_path_filter_edited(self):
        """
        Internal function that is called anytime the user edit the filter text
        Emits proper signal with the new filter text
        """

        if not self.emit_changes:
            return

        current_text = str(self.sub_path_filter.text()).strip()
        self.SubPathChanged.emit(current_text)

    def _on_sub_path_filter_changed(self):
        """
        Internal function that is called anytime filter text is changed
        Updates the directory used by the tree and updates the tree if necessary
        """

        current_text = str(self.sub_path_filter.text()).strip()
        if not current_text:
            self.set_directory(self.directory)
            if self.update_tree:
                self.tree_widget.set_directory(self.directory)
            text = self.filter_names.text()
            self._on_filter_names(text)
            return

        sub_dir = path.join_path(self.directory, current_text)
        if not sub_dir:
            return

        if path.is_dir(sub_dir):
            if self.update_tree:
                self.tree_widget.set_directory(sub_dir)
            text = self.filter_names.text()
            self._on_filter_names(text)

        if self.emit_changes:
            self.SubPathChanged.emit(current_text)

# ======================================================================================================


class FileTreeWidget(TreeWidget, object):

    refreshed = Signal()

    HEADER_LABELS = ['Name', 'Size MB', 'Time']
    NEW_ITEM_NAME = 'new_file'
    NEW_ITEM_WIDGET = QTreeWidgetItem
    EXCLUDE_EXTENSIONS = list()

    def __init__(self, parent=None):
        self._directory = None
        super(FileTreeWidget, self).__init__(parent)

        self.setHeaderLabels(self.HEADER_LABELS)

    def _add_sub_items(self, tree_item):
        """
        Implements _add_sub_items() functionality
        :param tree_item: QTreeWidgetItem
        """

        # Clean item hierarchy first
        self.delete_empty_children(tree_item)
        self.delete_tree_item_children(tree_item)

        path_str = self.get_tree_item_path_string(tree_item)
        full_path_str = path.join_path(self._directory, path_str)
        files = self._get_files(full_path_str)

        self._add_items(files, tree_item)

    def get_item_directory(self, tree_item):
        """
        Returns the full path of the given tree item
        :param tree_item: QTreeWidgetItem
        :return: str
        """

        path_str = self.get_tree_item_path_string(tree_item)
        return path.join_path(self._directory, path_str)

    def set_directory(self, directory, refresh=True):
        """
        Sets the directory used by this QTreeWidget
        :param directory: str, directory
        :param refresh: bool, Whether to refresh QTreeWidget items after setting working directory
        """

        qt.logger.debug('Setting Tree directory: {}'.format(directory))

        self._directory = directory
        if refresh:
            self.refresh()

    def create_item(self, name=None):
        """
        Creates a new item (folder) inside the selected item
        :param name: str, name of the new item
        """

        current_item = self._current_item
        if current_item:
            item_path = self.get_tree_item_path_string(self._current_item)
            item_path = path.join_path(self._directory, item_path)
            if path.is_file(item_path):
                item_path = path.get_dirname(item_path)
                current_item = self._current_item.parent()
        else:
            item_path = self._directory

        if not name:
            name = self.NEW_ITEM_NAME

        folder.create_folder(name=name, directory=item_path)

        if current_item:
            self._add_sub_items(current_item)
            self.setItemExpanded(current_item, True)
        else:
            self.refresh()

    def delete_item(self):
        """
        Deletes the selected item
        """

        item = self._current_item
        item_path = self.get_item_directory(item)
        name = path.get_basename(item_path)
        item_directory = path.get_dirname(item_path)

        if path.is_dir(item_path):
            folder.delete_folder(name, item_directory)
        elif path.is_file(item_path):
            fileio.delete_file(name, item_directory)
            if item_path.endswith('.py'):
                fileio.delete_file(name + '.c', item_directory)

        parent = item.parent()
        if parent:
            parent.removeChild(item)
        else:
            index = self.indexOfTopLevelItem(item)
            self.takeTopLevelItem(index)

    def refresh(self):
        """
        Refreshes all QTreeWidget items
        """

        if not self._directory:
            self.clear()
            return

        files = self._get_files()
        if not files:
            self.clear()
            return

        self._load_files(files)
        self.refreshed.emit()

    def _get_files(self, directory=None):
        """
        Internal function taht returns  all files located in the given directory. If not directory is given, stored
         variable directory will be used
        :return: list<str>
        """

        if directory is None:
            directory = self._directory

        return folder.get_files_and_folders(directory)

    def _load_files(self, files):
        """
        Internal function that adds given files into the tree (clearing the tree first)
        :param files: list<str>
        """

        self.clear()
        self._add_items(files)

    def _add_items(self, files, parent=None):
        """
        Internal function that adds given files into the tree
        :param files: list<str>, list of files to add
        :param parent: QTreeWidgetItem, parent item to append new item
        :return:
        """

        for f in files:
            if parent:
                self._add_item(f, parent)
            else:
                self._add_item(f)

    def _add_item(self, file_name, parent=None):
        """
        Internal function that adds given file into the tree
        :param file_name: str, name of the file new item will store
        :param parent: QTreeWidgetItem, parent item to append new item into
        :return: QTreeWidet, new item added
        """

        try:
            self.blockSignals(True)
            self.clearSelection()
        finally:
            self.blockSignals(False)

        path_name = file_name
        found = False

        # Check if item exists
        if parent:
            parent_path = self.get_tree_item_path_string(parent)
            path_name = '{}/{}'.format(parent_path, file_name)
            for i in range(parent.childCount()):
                item = parent.child(i)
                if item.text(0) == file_name:
                    found = item
        else:
            for i in range(self.topLevelItemCount()):
                item = self.topLevelItem(i)
                if item.text(0) == file_name:
                    found = item

        # Check if the item should be excluded or not from the tree
        exclude = self.EXCLUDE_EXTENSIONS
        if exclude:
            split_name = file_name.split('.')
            extension = split_name[-1]
            if extension in exclude:
                return

        if found:
            item = found
        else:
            item = self.ITEM_WIDGET()

        # Constrain item size if necessary
        size = self.ITEM_WIDGET_SIZE
        if size:
            size = QSize(*size)
            item.setSizeHint(self._title_text_index, size)

        # Set item text
        item_path = path.join_path(self._directory, path_name)
        sub_files = folder.get_files_and_folders(item_path)
        item.setText(self._title_text_index, file_name)

        # Retrieve file properties
        if self.header().count() > 1:
            if path.is_file(item_path):
                size = fileio.get_file_size(item_path)
                date = fileio.get_last_modified_date(item_path)
                item.setText(self._title_text_index + 1, str(size))
                item.setText(self._title_text_index + 2, str(date))

        # Update valid sub files
        # NOTE: Sub files are added dynamically when the user expands an item
        if sub_files:
            self.delete_tree_item_children(item)
            exclude_extensions = self.EXCLUDE_EXTENSIONS
            exclude_count = 0
            if exclude_extensions:
                for f in sub_files:
                    for exclude in exclude_extensions:
                        if f.endswith(exclude):
                            exclude_count += 1
                            break

            if exclude_count != len(sub_files):
                QTreeWidgetItem(item)

        # Add item to tree hierarchy
        if parent:
            parent.addChild(item)
            try:
                self.blockSignals(True)
                self.setCurrentItem(item)
            finally:
                self.blockSignals(False)
        else:
            self.addTopLevelItem(item)

        return item

    def get_tree_widget(self):
        """
        Returns TreeWidget being used by this widget
        Overrides to use custom TreeWidgets
        :return: QTreeWidget
        """

        return FileTreeWidget()

    def get_manager_widget(self):
        """
        Returns the widet that manages the tree widget
        :return: ManageTreeWidget
        """

        return ManageTreeWidget()

    def get_filter_widget(self):
        """
        Returns the filter widget used to filter tree contents
        :return: FilterTreeWidget
        """

        return FilterTreeDirectoryWidget()


class EditFileTreeWidget(base.DirectoryWidget, object):

    itemClicked = Signal(object, object)
    description = 'EditTree'

    TREE_WIDGET = FileTreeWidget
    MANAGER_WIDGET = ManageTreeWidget
    FILTER_WIDGET = FilterTreeWidget

    def __init__(self, parent=None):
        super(EditFileTreeWidget, self).__init__(parent)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

    def ui(self):
        super(EditFileTreeWidget, self).ui()

        self.tree_widget = self.TREE_WIDGET()

        self.manager_widget = self.MANAGER_WIDGET()
        self.manager_widget.set_tree_widget(self.tree_widget)

        self.filter_widget = self.FILTER_WIDGET()
        self.filter_widget.set_tree_widget(self.tree_widget)
        self.filter_widget.set_directory(self.directory)
        drag_reorder_icon = tp.ResourcesMgr().icon('drag_reorder')
        self.edit_mode_btn = buttons.IconButton(
            icon=drag_reorder_icon, icon_padding=2, button_style=buttons.ButtonStyles.FlatStyle)
        self.edit_mode_btn.setCheckable(True)
        self.edit_mode_btn.setMaximumHeight(20)
        self.edit_mode_btn.setMaximumWidth(40)
        self.filter_widget.main_layout.addWidget(self.edit_mode_btn)

        self.main_layout.addWidget(self.filter_widget)
        self.main_layout.addWidget(self.tree_widget)
        self.main_layout.addWidget(self.manager_widget)

        self._on_edit(False)

    def setup_signals(self):
        self.edit_mode_btn.toggled.connect(self._on_edit)
        self.tree_widget.itemClicked.connect(self._on_item_selection_changed)

    def set_directory(self, directory):
        """
        Overrides set_directory function to take in account also the sub path
        and update all directories of the different tree widgets
        :param directory: str
        :param sub_path: str
        """

        super(EditFileTreeWidget, self).set_directory(directory)
        self.filter_widget.set_directory(directory)
        self.tree_widget.set_directory(directory)

        if hasattr(self.manager_widget, 'set_directory'):
            self.manager_widget.set_directory(directory)

    def get_current_item(self):
        """
        Returns the current selected item on the tree
        :return: TreeItem
        """

        return self.tree_widget.current_item

    def get_current_item_name(self):
        """
        Returns the current name of the selected item on the tree
        :return: str
        """

        return self.tree_widget.current_name

    def get_current_item_directory(self):
        """
        Returns the directory the current selected item points to
        :return: str
        """

        item = self.get_current_item()
        return self.tree_widget.get_item_directory(item)

    def enable_edit_mode(self):
        """
        Enables edit mode
        """

        if not self.edit_mode_btn.isChecked():
            self.edit_mode_btn.blockSignals(True)
            try:
                self.edit_mode_btn.setChecked(True)
            finally:
                self.edit_mode_btn.blockSignals(False)

        self.tree_widget.setDragEnabled(True)
        self.tree_widget.setAcceptDrops(True)
        self.tree_widget.setDropIndicatorShown(True)

    def disable_edit_mode(self):
        """
        Enables edit mode
        """

        if self.edit_mode_btn.isChecked():
            self.edit_mode_btn.blockSignals(True)
            try:
                self.edit_mode_btn.setChecked(False)
            finally:
                self.edit_mode_btn.blockSignals(False)

        self.tree_widget.setDragEnabled(False)
        self.tree_widget.setAcceptDrops(False)
        self.tree_widget.setDropIndicatorShown(False)

    def get_checked_children(self, tree_item):
        """
        Function that returns checked item children of the given tree item
        :param tree_item:
        :return:
        """

        if not tree_item:
            return

        expand_state = tree_item.isExpanded()
        tree_item.setExpanded(True)
        children = self.tree_widget.get_tree_item_children(tree_item)

        checked_children = list()
        for child in children:
            check_state = child.checkState(0)
            if check_state == Qt.Checked:
                checked_children.append(child)
        levels = list()
        if checked_children:
            levels.append(checked_children)

        while children:
            new_children = list()
            checked_children = list()
            for child in children:
                current_check_state = child.checkState(0)
                if current_check_state != Qt.Checked:
                    continue
                child.setExpanded(True)
                sub_children = self.tree_widget.get_tree_item_children(child)
                checked = list()
                for sub_child in sub_children:
                    check_state = sub_child.checkState(0)
                    if check_state == Qt.Checked:
                        checked.append(sub_child)
                if sub_children:
                    new_children += sub_children
                if checked:
                    checked_children += checked
            if not checked_children:
                children = list()
                continue

        tree_item.setExpanded(expand_state)
        levels.reverse()

        return levels

    def refresh(self):
        """
        Refresh TreeWidget items
        """

        self.tree_widget.refresh()

    def _on_item_selection_changed(self):
        """
        Internal function that is called anytime the user selects an item on the TreeWidget
        Emits itemClicked signal with the name of the selected item and the item itself
        """

        items = self.tree_widget.selectedItems()
        name = None
        item = None
        if items:
            item = items[0]
            name = item.text(0)
            self.itemClicked.emit(name, item)

        return name, item

    def _on_edit(self, flag):
        """
        Internal function that is called anytime the user presses the Edit button on the filter widget
        If edit is ON, drag/drop operations in tree widget are disabled
        :param flag: bool
        """

        if flag:
            self.enable_edit_mode()
        else:
            self.disable_edit_mode()


class TreeWidgetItem(QTreeWidgetItem, object):
    def __init__(self, parent=None):
        self.widget = self.get_widget()
        if self.widget:
            self.widget.item = self
        self.column = self.get_column()
        super(TreeWidgetItem, self).__init__(parent)

    def get_widget(self):
        return None

    def get_column(self):
        return 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to show different kind of message boxes
"""

from __future__ import print_function, division, absolute_import

from Qt.QtWidgets import *

import tpDcc as tp


class MessageHandler(object):

    def __init__(self):
        super(MessageHandler, self).__init__()

        self.parent_window = tp.Dcc.get_main_window()

    def set_message(self, msg, level=0):
        """
        Sets a message in the status bar
        :param msg: str, message to show
        :param level: message level (0=info, 1=warning, 2=error)
        """

        if level < 0:
            level = 0
        if level > 3:
            level = 3

        # import maya.OpenMaya as OpenMaya
        # output = {0: OpenMaya.MGlobal.displayInfo, 1:
        # OpenMaya.MGlobal.displayWarning, 2: OpenMaya.MGlobal.displayError}
        # output[level](msg)

    def show_confirm_dialog(self, msg, title='Title'):
        """
        Shows a yes/no confirmation dialog
        :param msg: str, message to show with the dialog
        :param title: str, title of the dialog
        :return: bool, Whether the user has pressed yes(True) or No(False)
        """

        dialog = QMessageBox()
        dialog.setIcon(QMessageBox.Question)
        result = dialog.question(self.parent_window, title, msg, QMessageBox.Yes, QMessageBox.No)
        if result == QMessageBox.Yes:
            return True
        return False

    def show_warning_dialog(self, msg, detail=None):
        """
        Shows a warning dialog
        :param msg: str, message to show with the dialog
        :param detail: str, detail information to show (optional)
        """

        dialog = QMessageBox()
        dialog.setIcon(QMessageBox.Warning)
        dialog.setWindowTitle('Warning')
        dialog.setText(msg)
        if detail:
            dialog.setDetailedText(detail)
        dialog.exec_()

    def show_info_dialog(self, msg, title='Info'):
        """
        Shows a info dialog
        :param msg: str, message to show with the dialog
        :param title: str, title of the dialog
        """

        dialog = QMessageBox()
        dialog.setIcon(QMessageBox.Information)
        dialog.setWindowTitle(title)
        dialog.setText(msg)
        dialog.exec_()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widgets related with search functionality
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.widgets import buttons
from tpDcc.libs.qt.core import resource


class SearchFindWidget(QWidget, object):

    textChanged = Signal(str)
    editingFinished = Signal(str)
    returnPressed = Signal()

    def __init__(self, parent=None):
        super(SearchFindWidget, self).__init__(parent=parent)

        self.setObjectName('SearchFindWidget')

        self.text = ''
        self._placeholder_text = ''

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(2)
        self.setLayout(main_layout)

        self._search_line = QLineEdit(self)
        self._search_menu = QMenu()
        self._search_menu.addAction('Test')

        icon_size = self.style().pixelMetric(QStyle.PM_SmallIconSize)

        delete_icon = resource.Resource.icon('delete', extension='png')
        search_icon = resource.Resource.icon('search', extension='png')

        self._clear_btn = buttons.IconButton(delete_icon, icon_padding=2, parent=self)
        self._clear_btn.setIconSize(QSize(icon_size, icon_size))
        self._clear_btn.setFixedSize(QSize(icon_size, icon_size))
        self._clear_btn.hide()

        self._search_btn = buttons.IconButton(search_icon, icon_padding=2, parent=self)
        self._search_btn.setIconSize(QSize(icon_size, icon_size))
        self._search_btn.setFixedSize(QSize(icon_size, icon_size))
        # self._search_btn.setStyleSheet('border: none;')
        # self._search_btn.setPopupMode(QToolButton.InstantPopup)
        self._search_btn.setEnabled(True)

        self._search_line.setStyleSheet(
            """
            QLineEdit { padding-left: %spx; padding-right: %spx; border-radius:10px; border:2px; border-color:red; }
            """ % (self._search_button_padded_width(), self._clear_button_padded_width())
        )
        self._search_line.setMinimumSize(
            max(
                self._search_line.minimumSizeHint().width(),
                self._clear_button_padded_width() + self._search_button_padded_width()),
            max(
                self._search_line.minimumSizeHint().height(),
                max(self._clear_button_padded_width(), self._search_button_padded_width()))
        )

        main_layout.addWidget(self._search_line)

        self._search_line.setFocus()

        self._search_line.textChanged.connect(self.textChanged)
        self._search_line.textChanged.connect(self.set_text)
        # self._search_line.editingFinished.connect(self.editingFinished)
        # self._search_line.returnPressed.connect(self.returnPressed)
        self._clear_btn.clicked.connect(self.clear)
        self._search_btn.clicked.connect(self._popup_menu)

    # def get_text(self):
    #     return self._text
    #
    # def set_text(self, text):
    #     self._text = text
    #
    # text = property(get_text, set_text)

    def changeEvent(self, event):
        if event.type() == QEvent.EnabledChange:
            enabled = self.isEnabled()
            self._search_btn.setEnabled(enabled and self._search_menu)
            self._search_line.setEnabled(enabled)
            self._clear_btn.setEnabled(enabled)
        super(SearchFindWidget, self).changeEvent(event)

    def resizeEvent(self, event):
        if not (self._clear_btn and self._search_line):
            return
        super(SearchFindWidget, self).resizeEvent(event)
        x = self.width() - self._clear_button_padded_width() * 0.85
        y = (self.height() - self._clear_btn.height()) * 0.5
        self._clear_btn.move(x - 3, y)
        self._search_btn.move(self._search_line_frame_width() * 2, (self.height() - self._search_btn.height()) * 0.5)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.clear()
        super(SearchFindWidget, self).keyPressEvent(event)

    def get_text(self):
        if not self._search_line:
            return ''
        return self._search_line.text()

    def set_text(self, text):
        if not (self._clear_btn and self._search_line):
            return

        self._clear_btn.setVisible(not (len(text) == 0))
        if text != self.get_text():
            self._search_line.setText(text)

    def get_placeholder_text(self):
        if not self._search_line:
            return ''

        return self._search_line.text()

    def set_placeholder_text(self, text):
        if not self._search_line:
            return
        self._search_line.setPlaceholderText(text)

    def get_menu(self):
        search_icon = resource.icon('search')
        self._search_btn.setIcon(search_icon)
        self._search_btn.setEnabled(self.isEnabled() and self._menu)

    def set_focus(self, reason=Qt.OtherFocusReason):
        if self._search_line:
            self._search_line.setFocus(reason)
        else:
            self.setFocus(Qt.OtherFocusReason)

    def clear(self):
        if not self._search_line:
            return
        self._search_line.clear()
        self.set_focus()

    def select_all(self):
        if not self._search_line:
            return
        self._search_line.selectAll()

    def _search_line_frame_width(self):
        return self._search_line.style().pixelMetric(QStyle.PM_DefaultFrameWidth)

    def _clear_button_padded_width(self):
        return self._clear_btn.width() + self._search_line_frame_width() * 2

    def _clear_button_padded_height(self):
        return self._clear_btn.height() + self._search_line_frame_width() * 2

    def _search_button_padded_width(self):
        return self._search_btn.width() + self._search_line_frame_width() * 2

    def _search_button_padded_height(self):
        return self._search_btn.height() + self._search_line_frame_width() * 2

    def _popup_menu(self):
        if self._search_menu:
            screen_rect = QApplication.desktop().availableGeometry(self._search_btn)
            size_hint = self._search_menu.sizeHint()
            rect = self._search_btn.rect()
            top_diff = rect.top() - size_hint.height()
            x = rect.right() - size_hint.width() if self._search_btn.isRightToLeft() else rect.left()
            y = rect.bottom() if self._search_btn.mapToGlobal(
                QPoint(0, rect.bottom())).y() + size_hint.height() <= screen_rect.height() else top_diff
            point = self._search_btn.mapToGlobal(QPoint(x, y))
            point.setX(max(screen_rect.left(), min(point.x(), screen_rect.right() - size_hint.width())))
            point.setY(point.y() + 1)
            print('pop up on {}'.format(point))
            self._search_menu.popup(point)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

"""
This module includes custom splitter widgets
"""

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc
from tpDcc.libs.qt.widgets import buttons


class CollapsibleSplitter(QSplitter, object):

    doExpand = Signal()
    doCollapse = Signal()

    def __init__(self, orientation=Qt.Vertical, parent=None):

        self._handle = None
        super(CollapsibleSplitter, self).__init__(parent)

        self.setHandleWidth(16)
        self.setOrientation(orientation)

    def createHandle(self):
        self._handle = CollapsibleSplitterHandle(self.orientation(), self)
        self._handle.buttonClicked.connect(self._on_update_splitter)
        return self._handle

    def _on_update_splitter(self, expand):
        if expand:
            self.setSizes([1, 1])
        else:
            self.setSizes([1, 0])

    def expand(self):
        self._handle.expand()
        self.doExpand.emit()

    def collapse(self):
        self._handle.collapse()
        self.doCollapse.emit()


class CollapsibleSplitterButton(buttons.HoverButton, object):

    clickedButton = Signal()

    def __init__(self, icon=None, hover_icon=None, pressed_icon=None, parent=None):
        super(CollapsibleSplitterButton, self).__init__(
            icon=icon, hover_icon=hover_icon, pressed_icon=pressed_icon, parent=parent
        )

        self._can_emit_signal = True

    def mouseMoveEvent(self, event):
        super(CollapsibleSplitterButton, self).mouseMoveEvent(event)
        self._can_emit_signal = False
        if self._mouse_pressed:
            if self._pressed_icon:
                self.setIcon(self._pressed_icon)
            else:
                self.setIcon(self._idle_icon)

    def mouseReleaseEvent(self, event):
        if self.rect().contains(event.pos()) and self._can_emit_signal:
            self.clickedButton.emit()
        self._can_emit_signal = True
        super(CollapsibleSplitterButton, self).mouseReleaseEvent(event)


class CollapsibleSplitterHandle(QSplitterHandle, object):

    buttonClicked = Signal(bool)

    def __init__(self, orientation, parent):
        super(CollapsibleSplitterHandle, self).__init__(orientation, parent)

        self._is_expanded = True

        expand_icon = tpDcc.ResourcesMgr().icon('back')
        expand_hover_icon = tpDcc.ResourcesMgr().icon('expand_hover')
        expand_pressed_icon = tpDcc.ResourcesMgr().icon('expand_pressed')
        self.expand_btn = CollapsibleSplitterButton(
            icon=expand_icon, hover_icon=expand_hover_icon, pressed_icon=expand_pressed_icon, parent=self
        )
        self.expand_btn.setFixedWidth(16)
        self.expand_btn.setFixedHeight(16)
        self.expand_btn.setIconSize(QSize(16, 16))
        self.expand_btn.setStyleSheet('QWidget {background-color: rgba(255, 255, 255, 0); border:0px;}')
        self.expand_btn.setFocusPolicy(Qt.NoFocus)
        self.expand_btn.setVisible(False)

        collapse_icon = tpDcc.ResourcesMgr().icon('next')
        collapse_hover_icon = tpDcc.ResourcesMgr().icon('collapse_hover')
        collapse_pressed_icon = tpDcc.ResourcesMgr().icon('collapse_pressed')
        self.collapse_btn = CollapsibleSplitterButton(
            icon=collapse_icon, hover_icon=collapse_hover_icon, pressed_icon=collapse_pressed_icon, parent=self
        )
        self.collapse_btn.setFixedWidth(16)
        self.collapse_btn.setFixedHeight(16)
        self.collapse_btn.setIconSize(QSize(16, 16))
        self.collapse_btn.setStyleSheet('QWidget {background-color: rgba(255, 255, 255, 0); border:0px;}')
        self.collapse_btn.setFocusPolicy(Qt.NoFocus)
        self.collapse_btn.setVisible(False)
        if self.orientation() == Qt.Horizontal:
            self.collapse_btn.setCursor(Qt.SplitHCursor)
        else:
            self.collapse_btn.setCursor(Qt.SplitVCursor)

        self.expand_btn.clickedButton.connect(self.expand)
        self.collapse_btn.clickedButton.connect(self.collapse)

        self.expand()

    def mouseMoveEvent(self, event):
        if self._is_expanded:
            super(CollapsibleSplitterHandle, self).mouseMoveEvent(event)

    def resizeEvent(self, event):
        y = self.rect().height() * 0.5 - self.expand_btn.height() * 0.5
        self.expand_btn.move(0, y)
        self.collapse_btn.move(0, y)

    def paintEvent(self, event):

        painter = QPainter(self)

        if self._is_expanded:
            rect2 = event.rect()
            rect2.setX(15)
            rect2.setWidth(1)
            painter.fillRect(rect2, QBrush(QColor(87, 87, 87)))

    def collapse(self):
        self.expand_btn.setVisible(True)
        self.collapse_btn.setVisible(False)
        self._is_expanded = False
        self.setCursor(Qt.ArrowCursor)
        self.update()
        self.buttonClicked.emit(False)

    def expand(self):
        self.expand_btn.setVisible(False)
        self.collapse_btn.setVisible(True)
        self._is_expanded = True
        if self.orientation() == Qt.Horizontal:
            self.setCursor(Qt.SplitHCursor)
        else:
            self.setCursor(Qt.SplitVCursor)
        self.update()
        self.buttonClicked.emit(True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for custom PySide/PyQt windows
"""

from __future__ import print_function, division, absolute_import

import os
import uuid
from collections import defaultdict

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import appdirs

from tpDcc.libs import qt
import tpDcc as tp
from tpDcc import register
from tpDcc.libs.python import path, folder
from tpDcc.libs.qt.core import qtutils, animation, theme, statusbar, dragger, resizers, settings as qt_settings
from tpDcc.libs.qt.widgets import layouts


class WindowContents(QFrame, object):
    """
    Widget that defines the core contents of frameless window
    Can be used to custom CSS for frameless windows contents
    """

    def __init__(self, parent=None):
        super(WindowContents, self).__init__(parent=parent)


class BaseWindow(QMainWindow, object):

    closed = Signal()

    WindowName = 'New Window'

    def __init__(self, parent=None, **kwargs):

        main_window = tp.Dcc.get_main_window()
        if parent is None:
            parent = main_window

        window_id = kwargs.get('id', None)
        self._theme = None
        self._docks = list()
        self._toolbars = dict()
        self._dpi = kwargs.get('dpi', 1.0)
        self._fixed_size = kwargs.get('fixed_size', False)
        self._init_width = kwargs.get('width', 600)
        self._init_height = kwargs.get('height', 800)
        self._has_main_menu = False
        self._show_status_bar = kwargs.pop('show_statusbar', True)
        self._init_menubar = kwargs.pop('init_menubar', False)
        self._settings = kwargs.pop('settings', None)
        self._prefs_settings = kwargs.pop('preferences_settings', None)
        self._enable_save_position = True
        self._initial_pos_override = None
        self._signals = defaultdict(list)
        self._force_disable_saving = False
        win_settings = kwargs.pop('settings', None)
        prefs_settings = kwargs.pop('preferences_settings', None)
        auto_load = kwargs.get('auto_load', True)

        if not hasattr(self, 'WindowId'):
            if window_id:
                self.WindowId = window_id
            else:
                self._force_disable_saving = True
                self.WindowId = str(uuid.uuid4())

        super(BaseWindow, self).__init__(parent)

        self.setObjectName(str(self.WindowId))
        self.setWindowTitle(kwargs.get('title', self.WindowName))
        self.setWindowIcon(kwargs.get('icon', tp.ResourcesMgr().icon('tpdcc')))

        self.resize(self._init_width, self._init_height)
        self.center(self._init_width, self._init_height)

        # Load base generic window UI
        self._base_ui()

        # Load custom window UI
        self.ui()
        self.setup_signals()

        self.load_settings(settings=win_settings)

        if auto_load:
            self.load_theme()
        else:
            self.reload_stylesheet()

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def menuBar(self):
        return self._menubar

    def show(self, *args, **kwargs):
        """
        Shows the window and load its position
        :param self:
        :param args:
        :param kwargs:
        """

        super(BaseWindow, self).show()
        self.load_window_position()

        return self

    def closeEvent(self, event):
        self.save_settings()
        self.closed.emit()
        for child in self.findChildren(QWidget):
            child.close()
        self.setParent(None)
        self.deleteLater()

    def addDockWidget(self, area, dock_widget, orientation=Qt.Horizontal, tabify=True):
        """
        Overrides base QMainWindow addDockWidet function
        :param QDockWidgetArea area: area where dock will be added
        :param QDockWidget dock_widget: dock widget to add
        :param Qt.Orientation orientation: orientation fo the dock widget
        :param bool tabify: Whether or not dock widget can be tabbed
        """

        self._docks.append(dock_widget)
        if self._has_main_menu:
            self._view_menu.addAction(dock_widget.toggleViewAction())

        if tabify:
            for current_dock in self._docks:
                if self.dockWidgetArea(current_dock) == area:
                    self.tabifyDockWidget(current_dock, dock_widget)
                    dock_widget.setVisible(True)
                    dock_widget.setFocus()
                    dock_widget.raise_()
                    return

        super(BaseWindow, self).addDockWidget(area, dock_widget, orientation)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def process_events(self):
        """
        Forces Qt application to update GUI between calculations
        """

        QApplication.processEvents()

    def center(self, width=None, height=None):
        """
        Centers window to the center of the desktop
        :param width: int
        :param height: int
        """

        geometry = self.frameGeometry()
        if width:
            geometry.setWidth(width)
        if height:
            geometry.setHeight(height)

        desktop = QApplication.desktop()
        pos = desktop.cursor().pos()
        screen = desktop.screenNumber(pos)
        center_point = desktop.screenGeometry(screen).center()
        geometry.moveCenter(center_point)
        self.window().setGeometry(geometry)

    def center_to_parent(self, parent_geometry=None, child_geometry=None):
        """
        Centers current window to its parent
        :param parent_geometry:
        :param child_geometry:
        """

        if parent_geometry is None or child_geometry is None:
            base_window = self
            if parent_geometry is None:
                try:
                    parent_geometry = base_window.parent().frameGeometry()
                except AttributeError:
                    parent_geometry = QApplication.desktop().screenGeometry()
            if child_geometry is None:
                child_geometry = base_window.frameGeometry()

        self.move(
            parent_geometry.x() + (parent_geometry.width() - child_geometry.width()) / 2,
            parent_geometry.y() + (parent_geometry.height() - child_geometry.height()) / 2
        )

    def fade_close(self):
        """
        Closes the window with a fade animation
        """

        animation.fade_window(start=1, end=0, duration=400, object=self, on_finished=self.close)

    def show_ok_message(self, message, msecs=None):
        """
        Set an ok message to be displayed in the status bar
        :param message: str
        :param msecs: int
        """

        self._status_bar.show_ok_message(message=message, msecs=msecs)

    def show_info_message(self, message, msecs=None):
        """
        Set an info message to be displayed in the status bar
        :param message: str
        :param msecs: int
        """

        self._status_bar.show_info_message(message=message, msecs=msecs)

    def show_warning_message(self, message, msecs=None):
        """
       Set a warning message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        self._status_bar.show_warning_message(message=message, msecs=msecs)

    def show_error_message(self, message, msecs=None):
        """
       Set an error message to be displayed in the status widget
       :param message: str
       :param msecs: int
       """

        self._status_bar.show_error_message(message=message, msecs=msecs)

    # ============================================================================================================
    # UI
    # ============================================================================================================

    def get_main_layout(self):
        """
        Returns the main layout being used by the window
        :return: QLayout
        """

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(2)

        return main_layout

    def ui(self):
        """
        Function used to define UI of the window
        """

        pass

    def _base_ui(self):
        """
        Internal function that setup basic window UI
        """

        self.setDockNestingEnabled(True)
        self.setDocumentMode(True)
        self.setDockOptions(QMainWindow.AllowNestedDocks | QMainWindow.AnimatedDocks | QMainWindow.AllowTabbedDocks)
        self.setTabPosition(Qt.AllDockWidgetAreas, QTabWidget.North)

        # Central Widget

        central_widget = QWidget(parent=self)
        self.setCentralWidget(central_widget)
        self._central_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        central_widget.setLayout(self._central_layout)

        self._top_widget = QWidget()
        self._top_layout = layouts.VerticalLayout(margins=(0, 0, 0, 0), spacing=0)
        self._top_widget.setLayout(self._top_layout)

        # Status Bar
        #
        # self.statusBar().showMessage('')
        # self.statusBar().setSizeGripEnabled(not self._fixed_size)
        # self._status_bar = self.STATUS_BAR_WIDGET(self)
        # self.statusBar().addWidget(self._status_bar)
        # self.statusBar().setVisible(self._show_status_bar)

        # MenuBar
        self._menubar = QMenuBar()
        if self._init_menubar:
            self._has_main_menu = True
            self._file_menu = self.menuBar().addMenu('File')
            self._view_menu = self.menuBar().addMenu('View')
            self._exit_action = QAction(self)
            self._exit_action.setText('Close')
            self._exit_action.setShortcut('Ctrl + Q')
            self._exit_action.setIcon(tp.ResourcesMgr().resource.icon('close_window'))
            self._exit_action.setToolTip('Close application')
            self._file_menu.addAction(self._exit_action)
            self._exit_action.triggered.connect(self.fade_close)
            for i in self._docks:
                self._view_menu.addAction(i.toggleViewAction())
        self._top_layout.addWidget(self._menubar)

        self.main_widget = WindowContents()
        self.main_layout = self.get_main_layout()
        self.main_widget.setLayout(self.main_layout)

        self._central_layout.addWidget(self._top_widget)
        self._central_layout.addWidget(self.main_widget)

    # ============================================================================================================
    # SIGNALS
    # ============================================================================================================

    def setup_signals(self):
        """
        Override in derived class to setup signals
        This function is called after ui() function is called
        """

        pass

    def signal_connect(self, signal, fn, group=None):
        """
        Adds a new signal for the given group
        :param signal:
        :param fn:
        :param group:
        """

        self._signals[group].append((signal, fn))
        signal.connect(fn)

        return fn

    def signal_disconnect(self, group):
        """
        Disconnects and returns all functions for a current group
        :param group:
        :return: list
        """

        signals = list()
        for (signal, fn) in self._signals.pop(group, list()):
            try:
                signal.disconnect(fn)
            except RuntimeError:
                pass
            else:
                signals.append((signal, fn))

        return signals

    def signal_pause(self, *groups):
        """
        Pauses a certain set of signals during execution
        :param groups: list
        """

        if not groups:
            groups = self._signals

        signal_cache = dict()
        for group in groups:
            signal_cache[group] = self.signal_disconnect(group)

        yield

        for group in groups:
            for signal, fn in signal_cache[group]:
                self.signal_connect(signal, fn, group=group)

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns window settings
        :return: QtSettings
        """

        return self._settings

    def default_settings(self):
        """
        Returns default settings values
        :return: dict
        """

        return {}

    def set_settings(self, settings):
        """
        Set window settings
        :param settings:
        """

        self._settings = settings

        def_settings = self.default_settings()

        def_geometry = self.settings().get_default_value('geometry', self.objectName().upper())
        geometry = self.settings().getw('geometry', def_geometry)
        if geometry:
            self.restoreGeometry(geometry)

            # Reposition window in the center of the screen if the window is outside of the screen
            geometry = self.geometry()
            x = geometry.x()
            y = geometry.y()
            width = self._init_width or geometry.width()
            height = self._init_height or geometry.height()
            screen_geo = QApplication.desktop().screenGeometry()
            screen_width = screen_geo.width()
            screen_height = screen_geo.height()
            if x <= 0 or y <= 0 or x >= screen_width or y >= screen_height:
                self.center(width, height)

        def_window_state = self.settings().get_default_value('windowState', self.objectName().upper())
        window_state = self.settings().getw('windowState', def_window_state)
        if window_state:
            self.restoreState(window_state)

    def load_settings(self, settings=None):
        """
        Loads window settings from disk
        """

        settings = settings or self.settings()
        if not settings:
            settings = self._settings
            if not settings:
                self._settings = qt_settings.QtSettings(filename=self.get_settings_file(), window=self)
                self._settings.setFallbacksEnabled(False)
                if not self._prefs_settings:
                    self._prefs_settings = self._settings
            return self.set_settings(self._settings)

        return self.set_settings(settings)

    def save_settings(self, settings=None):
        """
        Saves window settings
        """

        settings = settings or self.settings()
        if not settings:
            return

        settings.setw('geometry', self.saveGeometry())
        settings.setw('saveState', self.saveState())
        settings.setw('windowState', self.saveState())

        return settings

    def get_settings_path(self):
        """
        Returns path where window settings are stored
        :return: str
        """

        return path.clean_path(os.path.join(appdirs.user_data_dir(), self.WindowId))

    def get_settings_file(self):
        """
        Returns file path of the window settings file
        :return: str
        """

        return path.clean_path(os.path.expandvars(os.path.join(self.get_settings_path(), 'settings.cfg')))

    def enable_save_window_position(self, enable):
        """
        Enables or disables the storage of window position in settings
        :param enable: bool
        """

        self._enable_save_position = enable

    def load_window_position(self):
        pass
        # if self._initial_pos_override is not None:
        #     x, y = self._initial_pos_override()
        #     x, y =

    def save_window_position(self):
        print('Saving window position ...')

    # ============================================================================================================
    # DPI
    # ============================================================================================================

    def dpi(self):
        """
        Return the current dpi for the window
        :return: float
        """

        return float(self._dpi)

    def set_dpi(self, dpi):
        """
        Sets current dpi for the window
        :param dpi: float
        """

        self._dpi = dpi

    # ============================================================================================================
    # THEME
    # ============================================================================================================

    def load_theme(self):
        """
        Loads window theme
        """

        def_settings = self.default_settings()
        def_theme_settings = def_settings.get('theme', dict())
        accent_color = self.settings().getw('theme/accent_color') or def_theme_settings.get('accent_color')
        background_color = self.settings().getw('theme/background_color') or def_theme_settings.get('background_color')
        accent_color = 'rgb(%d, %d, %d, %d)' % accent_color.getRgb() if isinstance(
            accent_color, QColor) else accent_color
        background_color = 'rgb(%d, %d, %d, %d)' % background_color.getRgb() if isinstance(
            background_color, QColor) else background_color

        theme_settings = dict()
        if accent_color:
            theme_settings['accent_color'] = accent_color
        if background_color:
            theme_settings['background_color'] = background_color

        self.set_theme_settings(theme_settings)

    def theme(self):
        """
        Returns the current theme
        :return: Theme
        """

        if not self._theme:
            return None

        return self._theme

    def set_theme(self, theme):
        """
        Sets current window theme
        :param theme: Theme
        """

        self._theme = theme
        self._theme.updated.connect(self.reload_stylesheet)
        self.reload_stylesheet()

    def set_theme_settings(self, settings):
        """
        Sets the theme settings from the given settings
        :param settings: dict
        """

        current_theme = self._settings.get('theme', 'default')
        new_theme = tp.ResourcesMgr().theme(current_theme)
        if not new_theme:
            new_theme = theme.Theme()
        new_theme.set_settings(settings)
        self.set_theme(new_theme)

    def reload_stylesheet(self):
        """
        Reloads the stylesheet to the current theme
        """

        current_theme = self.theme()
        current_theme.set_dpi(self.dpi())
        stylesheet = current_theme.stylesheet()
        self.setStyleSheet(stylesheet)

        # TODO: This operation is VERY heavy. Find a better way of doing this
        # all_widgets = qtutils.iterate_children(self.main_widget, qobj_class=QObject)
        # for w in all_widgets:
        #     if hasattr(w, 'setStyleSheet'):
        #         w.setStyleSheet(stylesheet)
        #         w.update()

    # ============================================================================================================
    # TOOLBAR
    # ============================================================================================================

    def add_toolbar(self, name, area=Qt.TopToolBarArea):
        """
        Adds a new toolbar to the window
        :return:  QToolBar
        """

        new_toolbar = QToolBar(name)
        self.addToolBar(area, new_toolbar)

        return new_toolbar

    # ============================================================================================================
    # DOCK
    # ============================================================================================================

    def add_dock(self, name, widget=None, pos=Qt.LeftDockWidgetArea, tabify=True):
        """
        Adds a new dockable widget to the window
        :param name: str, name of the dock widget
        :param widget: QWidget, widget to add to the dock
        :param pos: Qt.WidgetArea
        :param tabify: bool, Wheter the new widget should be tabbed to existing docks
        :return: QDockWidget
        """

        if widget:
            dock_name = ''.join([widget.objectName(), 'Dock'])
        else:
            dock_name = name + 'Dock'

        existing_dock = self.find_dock(dock_name)
        if existing_dock:
            existing_dock.raise_()

        dock = DockWidget(title=name, parent=self, floating=False)
        dock.setObjectName(dock_name)
        if widget is not None:
            dock.setWidget(widget)
        self.addDockWidget(pos, dock, tabify=tabify)

        return dock

    def set_active_dock_tab(self, dock_widget):
        """
        Sets the current active dock tab depending on the given dock widget
        :param dock_widget: DockWidget
        """

        tab_bars = self.findChildren(QTabBar)
        for bar in tab_bars:
            count = bar.count()
            for i in range(count):
                data = bar.tabData(i)
                widget = qtutils.to_qt_object(data, qobj=type(dock_widget))
                if widget == dock_widget:
                    bar.setCurrentIndex(i)

    def find_dock(self, dock_name):
        """
        Returns the dock widget based on the object name passed
        :param str dock_name: dock objectName to find
        :return: QDockWidget or None
        """

        for dock in self._docks:
            if dock.objectName() == dock_name:
                return dock

        return None

    def _parent_override(self):
        """
        Internal function that overrides parent functionality to make sure that proper parent attributes are used
        in dockable windows
        """

        # Make sure this function is inherited
        return super(MainWindow, self)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _load_ui_from_file(self, ui_file):
        """
        Internal function that loads given UI file
        :param ui_file: str
        :return: QWidget or None
        """

        if not os.path.isfile(ui_file):
            return None

        loaded_ui = qtutils.load_ui(ui_file=ui_file)

        return loaded_ui


class MainWindow(BaseWindow, object):
    """
    Main class to create windows
    """

    dockChanged = Signal(object)
    windowResizedFinished = Signal()
    framelessChanged = Signal(object)
    windowReady = Signal()
    clearedInstance = Signal()

    STATUS_BAR_WIDGET = statusbar.StatusWidget
    DRAGGER_CLASS = dragger.WindowDragger

    _WINDOW_INSTANCES = dict()

    def __init__(self, parent=None, **kwargs):

        self._setup_resizers()

        self._preference_widgets_classes = list()
        self._toolset = kwargs.get('toolset', None)
        self._transparent = kwargs.get('transparent', False)
        self._config = kwargs.pop('config', None)
        self._dockable = getattr(self, 'WindowDockable', False)
        self._was_docked = False
        self._window_loaded = False
        self._window_closed = False
        self._current_docked = None

        super(MainWindow, self).__init__(parent=parent, **kwargs)

        frameless = kwargs.get('frameless', True)
        self.set_frameless(frameless)
        if not frameless:
            self.set_resizer_active(False)
            self._dragger.set_dragging_enabled(False)
            self._dragger.set_window_buttons_state(False)
        else:
            self._dragger.set_dragging_enabled(True)
            self._dragger.set_window_buttons_state(True)
        self._dragger._toggle_frameless.setChecked(self.is_frameless())

        # We set the window title after UI is created
        self.setWindowTitle(kwargs.get('title', 'tpDcc'))
        self.setWindowIcon(kwargs.get('icon', tp.ResourcesMgr().icon('tpdcc')))

        MainWindow._WINDOW_INSTANCES[self.WindowId] = {
            'window': self
        }

        self.windowReady.connect(lambda: setattr(self, '_window_loaded', True))

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def widget(self):
        """
        Returns widget
        """

        return self._widget

    # ============================================================================================================
    # CLASS METHODS
    # ============================================================================================================

    @classmethod
    def instance(cls, parent=None, **kwargs):
        pass

    @classmethod
    def clear_window_instance(cls, window_id):
        """
        Closes the last class instance
        :param window_id:
        :return:
        """

        inst = cls._WINDOW_INSTANCES.pop(window_id, None)
        if inst is not None:
            try:
                inst['window'].clearedInstance.emit()
            except RuntimeError as exc:
                tp.logger.error('Error while clearing window instance: {} | {}'.format(window_id, exc))

        return inst

    @classmethod
    def clear_window_instances(cls):
        """
        Closes every loaded window
        """

        for window_id in tuple(cls._WINDOW_INSTANCES):
            cls.clear_window_instance(window_id)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def showEvent(self, event):
        if self.docked() != self._current_docked:
            self._current_docked = self.docked()
            self.dockChanged.emit(self._current_docked)

        super(MainWindow, self).showEvent(event)

    def closeEvent(self, event):
        self._window_closed = True
        self.unregister_callbacks()
        self.clear_window_instance(self.WindowId)
        super(MainWindow, self).closeEvent(event)

    def setWindowIcon(self, icon):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_icon(icon)
        super(MainWindow, self).setWindowIcon(icon)

    def setWindowTitle(self, title):
        if self.is_frameless() or (hasattr(self, '_dragger') and self._dragger):
            self._dragger.set_title(title)
        super(MainWindow, self).setWindowTitle(title)

    def show(self, *args, **kwargs):
        """
        Shows the window and load its position
        :param self:
        :param args:
        :param kwargs:
        """

        super(MainWindow, self).show()
        self.windowReady.emit()

        return self

    # ============================================================================================================
    # UI
    # ============================================================================================================

    def ui(self):
        """
        Function used to define UI of the window
        """

        super(MainWindow, self).ui()

        for r in self._resizers:
            r.setParent(self)

        # Dragger
        self._dragger = self.DRAGGER_CLASS(window=self)
        self._top_layout.insertWidget(0, self._dragger)

        for r in self._resizers:
            r.windowResizedFinished.connect(self.windowResizedFinished)
        self.set_resize_directions()

        grid_layout = layouts.GridLayout()
        grid_layout.setHorizontalSpacing(0)
        grid_layout.setVerticalSpacing(0)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.addWidget(self._top_widget, 1, 1, 1, 1)
        grid_layout.addWidget(self.main_widget, 2, 1, 1, 1)
        grid_layout.addWidget(self._top_left_resizer, 0, 0, 1, 1)
        grid_layout.addWidget(self._top_resizer, 0, 1, 1, 1)
        grid_layout.addWidget(self._top_right_resizer, 0, 2, 1, 1)
        grid_layout.addWidget(self._left_resizer, 1, 0, 2, 1)
        grid_layout.addWidget(self._right_resizer, 1, 2, 2, 1)
        grid_layout.addWidget(self._bottom_left_resizer, 3, 0, 1, 1)
        grid_layout.addWidget(self._bottom_resizer, 3, 1, 1, 1)
        grid_layout.addWidget(self._bottom_right_resizer, 3, 2, 1, 1)
        grid_layout.setColumnStretch(1, 1)
        grid_layout.setRowStretch(2, 1)

        self._central_layout.addLayout(grid_layout)

        # Shadow effect for window
        # BUG: This causes some rendering problems when using other shadow effects in child widgets of the window
        # BUG: Also detected problems when updating wigets (tree views, web browsers, etc)
        # https://bugreports.qt.io/browse/QTBUG-35196
        # shadow_effect = QGraphicsDropShadowEffect(self)
        # shadow_effect.setBlurRadius(qtutils.dpi_scale(15))
        # shadow_effect.setColor(QColor(0, 0, 0, 150))
        # shadow_effect.setOffset(qtutils.dpi_scale(0))
        # self.setGraphicsEffect(shadow_effect)

        for r in self._resizers:
            r.windowResizedFinished.connect(self.windowResizedFinished)

        if self._toolset:
            self.main_layout.addWidget(self._toolset)

    # ============================================================================================================
    # SIGNALS
    # ============================================================================================================

    def register_callback(self, callback_type, fn):
        """
        Registers the given callback with the given function
        :param callback_type: tpDcc.DccCallbacks
        :param fn: Python function to be called when callback is emitted
        """

        if type(callback_type) in [list, tuple]:
            callback_type = callback_type[0]

        if callback_type not in tp.callbacks():
            tp.logger.warning('Callback Type: "{}" is not valid! Aborting callback creation ...'.format(callback_type))
            return

        from tpDcc.managers import callbacks
        return callbacks.CallbacksManager.register(callback_type=callback_type, fn=fn, owner=self)

    def unregister_callbacks(self):
        """
        Unregisters all callbacks registered by this window
        """

        from tpDcc.managers import callbacks
        callbacks.CallbacksManager.unregister_owner_callbacks(owner=self)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def exists(self):
        """
        Returns whether or not this window exists
        :return: bool
        """

        return True

    def is_loaded(self):
        """
        Returns whether or not this window has been already loaded
        :return: bool
        """

        return self._window_loaded and not self.is_closed()

    def is_closed(self):
        """
        Returns whether or not this window has been closed
        """

        return self._window_closed

    def is_frameless(self):
        """
        Returns whether or not frameless functionality for this window is enable or not
        :return: bool
        """

        return self.window().windowFlags() & Qt.FramelessWindowHint == Qt.FramelessWindowHint

    def set_frameless(self, flag):
        """
        Sets whether frameless functionality is enabled or not
        :param flag: bool
        :param show: bool
        """

        window = self.window()

        if flag and not self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
                window.setWindowFlags(window.windowFlags() ^ Qt.WindowMinMaxButtonsHint)
            else:
                window.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
                window.setWindowFlags(window.windowFlags() ^ Qt.WindowMinMaxButtonsHint)
            self.set_resizer_active(True)
        elif not flag and self.is_frameless():
            window.setAttribute(Qt.WA_TranslucentBackground)
            if qtutils.is_pyside2() or qtutils.is_pyqt5():
                window.setWindowFlags(window.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            else:
                self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
            self.set_resizer_active(False)

        window.show()

    # ============================================================================================================
    # RESIZERS
    # ============================================================================================================

    def set_resizer_active(self, flag):
        """
        Sets whether resizers are enable or not
        :param flag: bool
        """

        if flag:
            for r in self._resizers:
                r.show()
        else:
            for r in self._resizers:
                r.hide()

    def set_resize_directions(self):
        """
        Sets the resize directions for the resizer widget of this window
        """

        self._top_resizer.set_resize_direction(resizers.ResizeDirection.Top)
        self._bottom_resizer.set_resize_direction(resizers.ResizeDirection.Bottom)
        self._right_resizer.set_resize_direction(resizers.ResizeDirection.Right)
        self._left_resizer.set_resize_direction(resizers.ResizeDirection.Left)
        self._top_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Top)
        self._top_right_resizer.set_resize_direction(resizers.ResizeDirection.Right | resizers.ResizeDirection.Top)
        self._bottom_left_resizer.set_resize_direction(resizers.ResizeDirection.Left | resizers.ResizeDirection.Bottom)
        self._bottom_right_resizer.set_resize_direction(
            resizers.ResizeDirection.Right | resizers.ResizeDirection.Bottom)

    def get_resizers_height(self):
        """
        Returns the total height of the vertical resizers
        :return: float
        """

        resizers = [self._top_resizer, self._bottom_resizer]
        total_height = 0
        for r in resizers:
            if not r.isHidden():
                total_height += r.minimumSize().height()

        return total_height

    def get_resizers_width(self):
        """
        Returns the total widht of the horizontal resizers
        :return: float
        """

        resizers = [self._left_resizer, self._right_resizer]
        total_width = 0
        for r in resizers:
            if not r.isHidden():
                total_width += r.minimumSize().width()

        return total_width

    def _setup_resizers(self):
        """
        Internal function that setup window resizers
        """

        self._top_resizer = resizers.VerticalResizer()
        self._bottom_resizer = resizers.VerticalResizer()
        self._right_resizer = resizers.HorizontalResizer()
        self._left_resizer = resizers.HorizontalResizer()
        self._top_left_resizer = resizers.CornerResizer()
        self._top_right_resizer = resizers.CornerResizer()
        self._bottom_left_resizer = resizers.CornerResizer()
        self._bottom_right_resizer = resizers.CornerResizer()

        self._resizers = [
            self._top_resizer, self._top_right_resizer, self._right_resizer, self._bottom_right_resizer,
            self._bottom_resizer, self._bottom_left_resizer, self._left_resizer, self._top_left_resizer
        ]

    # ============================================================================================================
    # PREFERENCES SETTINGS
    # ============================================================================================================

    def preferences_settings(self):
        """
        Returns window preferences settings
        :return: QtSettings
        """

        return self._prefs_settings

    def set_preferences_settings(self, prefs_settings):
        """
        Sets window preference settings
        :param prefs_settings:
        """

        self._prefs_settings = prefs_settings

    def register_preference_widget_class(self, widget_class):
        """
        Function used to registere preference widgets
        """

        if not hasattr(widget_class, 'CATEGORY'):
            qt.logger.warning(
                'Impossible to register Category Wigdet Class "{}" because it does not '
                'defines a CATEGORY attribute'.format(widget_class))
            return

        registered_prefs_categories = [pref.CATEGORY for pref in self._preference_widgets_classes]
        if widget_class.CATEGORY in registered_prefs_categories:
            qt.logger.warning(
                'Impossible to register Category Widget Class "{}" because its CATEGORY "{}" its '
                'already registered!'.format(widget_class, widget_class.CATEGORY))
            return

        self._preference_widgets_classes.append(widget_class)

    # ============================================================================================================
    # DOCK
    # ============================================================================================================

    def dockable(self, raw=False):
        """
        Returns whether or not the window is dockable
        :param raw: bool, If True, get current state of the window, otherwise get current setting
        :return: bool
        """

        if not raw and self._was_docked is not None:
            return self._was_docked

        return self._dockable

    def set_dockable(self, dockable, override=False):
        """
        Sets whether or not this window is dockable
        :param dockable: bool
        :param override: bool, If the dockable raw value should be set.
            Only should be used if the dock state has changed
        """

        if override:
            self._was_docked = self._dockable = dockable
        else:
            self._was_docked = self._dockable
            self._dockable = dockable
            self.save_window_position()

    def docked(self):
        """
        Returns whether or not this window is currently docked
        :return: bool
        """

        if not self.dockable():
            return False

        raise NotImplementedError('docked function is not implemented!')

    def is_floating(self):
        """
        Returns whether or not this window is floating
        :return: bool
        """

        return tp.Dcc.is_window_floating(self.WindowId)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _settings_validator(self, **kwargs):
        """
        Validator used for the settings dialog
        :param kwargs: dict
        """

        fields = list()

        clr = kwargs.get("accentColor")
        if clr and self.theme().accent_color().to_string() != clr:
            self.theme().set_accent_color(clr)

        clr = kwargs.get("backgroundColor")
        if clr and self.theme().background_color().to_string() != clr:
            self.theme().set_background_color(clr)

        return fields

    def _settings_accepted(self, **kwargs):
        """
        Function that is called when window settings dialog are accepted
        :param kwargs: dict
        """

        if not self.settings():
            return

        theme_name = self.theme().name()
        accent_color = kwargs.get('accentColor', self.theme().accent_color().to_string())
        background_color = kwargs.get('backgroundColor', self.theme().background_color().to_string())
        if theme_name:
            self.settings().setw('theme/name', theme_name)
        self.settings().setw('theme/accentColor', accent_color)
        self.settings().setw('theme/backgroundColor', background_color)
        self.settings().sync()

        self.load_theme()

    def _setup_theme_preferences(self):

        from tpDcc.libs.qt.core import preferences
        from tpDcc.libs.qt.widgets import formwidget

        accent_color = self.theme().accent_color().to_string()
        background_color = self.theme().background_color().to_string()
        settings_validator = self._settings_validator
        settings_accepted = self._settings_accepted

        class ThemeCategoryWidget(preferences.CategoryWidgetBase, object):

            CATEGORY = 'Theme'

            def __init__(self, parent=None):
                super(ThemeCategoryWidget, self).__init__(parent=parent)

                self.main_layout = QVBoxLayout()
                self.main_layout.setContentsMargins(2, 2, 2, 2)
                self.main_layout.setSpacing(2)
                self.setLayout(self.main_layout)

                form = {
                    "title": "Theme",
                    "description": "Theme Colors",
                    "layout": "vertical",
                    "schema": [
                        {
                            "name": "accentColor",
                            "type": "color",
                            "value": accent_color,
                            "colors": [
                                "rgb(230, 80, 80, 255)",
                                "rgb(230, 125, 100, 255)",
                                "rgb(230, 120, 40)",
                                "rgb(240, 180, 0, 255)",
                                "rgb(80, 200, 140, 255)",
                                "rgb(50, 180, 240, 255)",
                                "rgb(110, 110, 240, 255)",
                            ]
                        },
                        {
                            "name": "backgroundColor",
                            "type": "color",
                            "value": background_color,
                            "colors": [
                                "rgb(40, 40, 40)",
                                "rgb(68, 68, 68)",
                                "rgb(80, 60, 80)",
                                "rgb(85, 60, 60)",
                                "rgb(60, 75, 75)",
                                "rgb(60, 64, 79)",
                                "rgb(245, 245, 255)",
                            ]
                        },
                    ],
                    "validator": settings_validator,
                    "accepted": settings_accepted
                }

                self._dlg = formwidget.FormDialog(parent=parent, form=form)
                self._dlg.setMinimumWidth(300)
                self._dlg.setMinimumHeight(300)
                self._dlg.setMaximumWidth(400)
                self._dlg.setMaximumHeight(400)
                self._dlg.accept_button().setText('Save')
                self._dlg.accept_button().setVisible(False)
                self._dlg.reject_button().setVisible(False)
                self._dlg.show()
                self.main_layout.addWidget(self._dlg)

        theme_prefs_widget = ThemeCategoryWidget(parent=self._preferences_window)

        return theme_prefs_widget


class DetachedWindow(QMainWindow):
    """
    Class that incorporates functionality to create detached windows
    """

    windowClosed = Signal(object)

    class DetachPanel(QWidget, object):
        widgetVisible = Signal(QWidget, bool)

        def __init__(self, parent=None):
            super(DetachedWindow.DetachPanel, self).__init__(parent=parent)

            self.main_layout = QVBoxLayout()
            self.setLayout(self.main_layout)

        def set_widget_visible(self, widget, visible):
            self.setVisible(visible)
            self.widgetVisible.emit(widget, visible)

        def set_widget(self, widget):
            qtutils.clear_layout(self.main_layout)
            self.main_layout.addWidget(widget)
            widget.show()

    class SettingGroup(object):
        global_group = ''

        def __init__(self, name):
            self.name = name
            self.settings = QSettings()

        def __enter__(self):
            if self.global_group:
                self.settings.beginGroup(self.global_group)
            self.settings.beginGroup(self.name)
            return self.settings

        def __exit__(self, *args):
            if self.global_group:
                self.settings.endGroup()
            self.settings.endGroup()
            self.settings.sync()

        @staticmethod
        def load_basic_window_settings(window, window_settings):
            window.restoreGeometry(window_settings.value('geometry', QByteArray()))
            window.restoreState(window_settings.value('windowstate', QByteArray()))
            try:
                window.split_state = window_settings.value('splitstate', '')
            except TypeError:
                window.split_state = ''

    def __init__(self, title, parent):
        self.tab_idx = -1
        super(DetachedWindow, self).__init__(parent=parent)

        self.main_widget = self.DetachPanel()
        self.setCentralWidget(self.main_widget)

        self.setWindowTitle(title)
        self.setWindowModality(Qt.NonModal)
        self.sgroup = self.SettingGroup(title)
        with self.sgroup as config:
            self.SettingGroup.load_basic_window_settings(self, config)

        self.statusBar().hide()

    def closeEvent(self, event):
        with self.sgroup as config:
            config.setValue('detached', False)
        self.windowClosed.emit(self)
        self.deleteLater()

    def moveEvent(self, event):
        super(DetachedWindow, self).moveEvent(event)
        self.save_settings()

    def resizeEvent(self, event):
        super(DetachedWindow, self).resizeEvent(event)
        self.save_settings()

    def set_widget_visible(self, widget, visible):
        self.setVisible(visible)

    def set_widget(self, widget):
        self.main_widget.set_widget(widget=widget)

    def save_settings(self, detached=True):
        with self.sgroup as config:
            config.setValue('detached', detached)
            config.setValue('geometry', self.saveGeometry())
            config.setValue('windowstate', self.saveState())


class DockWindow(QMainWindow, object):
    """
    Class that with dock functionality. It's not intended to use as main window (use MainWindow for that) but for
    being inserted inside a window and have a widget with dock functionality in the main layout of that window
    """

    class DockWidget(QDockWidget, object):
        def __init__(self, name, parent=None, window=None):
            super(DockWindow.DockWidget, self).__init__(name, parent)

            self.setWidget(window)

        # region Override Functions
        def setWidget(self, widget):
            """
            Sets the window instance of the dockable main window
            """

            super(DockWindow.DockWidget, self).setWidget(widget)

            if widget and issubclass(widget.__class__, MainWindow):
                # self.setFloating(True)
                self.setWindowTitle(widget.windowTitle())
                self.visibilityChanged.connect(self._visibility_changed)

                widget.setWindowFlags(Qt.Widget)
                widget.setParent(self)
                widget.windowTitleChanged.connect(self._window_title_changed)

        # endregion

        # region Private Functions
        def _visibility_changed(self, state):
            """
            Process QDockWidget's visibilityChanged signal
            """

            # TODO: Implement export widget properties functionality
            # widget = self.widget()
            # if widget:
            #     widget.export_settings()

        def _window_title_changed(self, title):
            """
            Process BaseWindow's windowTitleChanged signal
            :param title: str, new title
            """

            self.setWindowTitle(title)

    _last_instance = None

    def __init__(self, name='BaseWindow', title='DockWindow', use_scrollbar=False, parent=None):
        self.main_layout = self.get_main_layout()
        self.__class__._last_instance = self
        super(DockWindow, self).__init__(parent)

        self.docks = list()
        self.connect_tab_change = True
        self.use_scrollbar = use_scrollbar

        self.setObjectName(name)
        self.setWindowTitle(title)
        self.statusBar().setSizeGripEnabled(False)
        self.statusBar().hide()

        self.ui()

        self.tab_change_hide_show = True

    def keyPressEvent(self, event):
        return

    def get_main_layout(self):
        """
        Function that generates the main layout used by the widget
        Override if necessary on new widgets
        :return: QLayout
        """

        return QVBoxLayout()

    def ui(self):
        """
        Function that sets up the ui of the widget
        Override it on new widgets (but always call super)
        """

        main_widget = QWidget()
        if self.use_scrollbar:
            scroll = QScrollArea()
            scroll.setWidgetResizable(True)
            scroll.setWidget(main_widget)
            self._scroll_widget = scroll
            main_widget.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
            self.setCentralWidget(scroll)
        else:
            self.setCentralWidget(main_widget)

        main_widget.setLayout(self.main_layout)
        self.main_widget = main_widget

        self.main_layout.expandingDirections()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        # ==========================================================================================

        # TODO: Check if we should put this on constructor
        # self.main_widget.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum))
        # self.centralWidget().hide()

        self.setTabPosition(Qt.TopDockWidgetArea, QTabWidget.West)
        self.setDockOptions(self.AnimatedDocks | self.AllowTabbedDocks | self.AllowNestedDocks)

    def set_active_dock_tab(self, dock_widget):
        """
        Sets the current active dock tab depending on the given dock widget
        :param dock_widget: DockWidget
        """

        tab_bars = self.findChildren(QTabBar)
        for bar in tab_bars:
            count = bar.count()
            for i in range(count):
                data = bar.tabData(i)
                widget = qtutils.to_qt_object(data, qobj=type(dock_widget))
                if widget == dock_widget:
                    bar.setCurrentIndex(i)

    def add_dock(self, widget, name, pos=Qt.TopDockWidgetArea, tabify=True):
        docks = self._get_dock_widgets()
        for dock in docks:
            if dock.windowTitle() == name:
                dock.deleteLater()
                dock.close()
        dock_widget = self.DockWidget(name=name, parent=self)
        # dock_widget.setSizePolicy(QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum))
        dock_widget.setAllowedAreas(pos)
        dock_widget.setWidget(widget)

        self.addDockWidget(pos, dock_widget)

        if docks and tabify:
            self.tabifyDockWidget(docks[-1], dock_widget)

        dock_widget.show()
        dock_widget.raise_()

        tab_bar = self._get_tab_bar()
        if tab_bar:
            if self.connect_tab_change:
                tab_bar.currentChanged.connect(self._on_tab_changed)
                self.connect_tab_change = False

        return dock_widget

    def _get_tab_bar(self):
        children = self.children()
        for child in children:
            if isinstance(child, QTabBar):
                return child

    def _get_dock_widgets(self):
        found = list()
        for child in self.children():
            if isinstance(child, QDockWidget):
                found.append(child)

        return found

    def _on_tab_changed(self, index):
        if not self.tab_change_hide_show:
            return

        docks = self._get_dock_widgets()

        docks[index].hide()
        docks[index].show()


class SubWindow(MainWindow, object):
    """
    Class to create sub windows
    """

    def __init__(self, parent=None, **kwargs):
        super(SubWindow, self).__init__(parent=parent, frameless=False, **kwargs)


class DirectoryWindow(MainWindow, object):
    """
    Window that stores variable to store current working directory
    """

    def __init__(self, parent=None, **kwargs):
        self.directory = None
        super(DirectoryWindow, self).__init__(parent=parent, frameless=False, **kwargs)

    def set_directory(self, directory):
        """
        Sets the directory of the window. If the given folder does not exists, it will created automatically
        :param directory: str, new directory of the window
        """

        self.directory = directory

        if not path.is_dir(directory=directory):
            folder.create_folder(name=None, directory=directory)


class DockWidget(QDockWidget, object):
    """
    Base docked widget
    """

    def __init__(self, title, parent=None, floating=False):
        super(DockWidget, self).__init__(title, parent)

        self.setFloating(floating)
        self.setFeatures(
            QDockWidget.DockWidgetMovable | QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetClosable)


class DockWindowContainer(DockWidget, object):
    """
    Docked Widget used to dock windows inside other windows
    """

    def __init__(self, title):
        super(DockWindowContainer, self).__init__(title)

    def closeEvent(self, event):
        if self.widget():
            self.widget().close()
        super(DockWindowContainer, self).closeEvent(event)


register.register_class('Window', MainWindow)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different grid widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtWidgets import *

from tpDcc.libs.qt.core import base


class GridWidget(QTableWidget, object):
    """
    Widget that behaves as a grid widget
    """

    def __init__(self, parent=None):
        super(GridWidget, self).__init__(parent=parent)

    def resizeEvent(self, event):
        """
        Override resizeEvent so the table columns maintain its size when resizing QTableWidget
        :param event: QResizeEvent
        :return:
        """
        total_width = self.viewport().width()
        num_columns = self.columnCount()
        for column in range(num_columns):
            self.horizontalHeader().resizeSection(column, total_width / num_columns)

    def pos_to_row_col(self, pos):
        """
        With a given position in the grid returns the (row, column) values for that position
        :param pos: QPoint
        :return:
        """
        index = self.indexAt(pos)
        if not index:
            return None
        return (index.row(), index.column())

    def addWidget(self, row, col, widget):
        """
        Adds a new QWidget into the given row col overriding any existing widget if already exists one
        :param row: int, row to add the widget into
        :param col: int, col to add the widget into
        :param widget: QWidget
        """

        container_widget = base.ContainerWidget(self)
        container_widget.set_contained_widget(widget)
        self.setCellWidget(row, col, container_widget)

    def add_widget_first_empty_cell(self, widget):
        """
        Adds a new QWidget into the first available cell in the grid
        :param widget: QWidget
        :return:
        """

        row, col = self.first_empty_cell()
        self.addWidget(row, col, widget)
        self.resizeRowsToContents()

    def first_empty_cell(self):
        """
        Returns the first empty cell in the table
        :return: int, int, first row, col pair of empty the first empty cell
        """

        empty_cell = (-1, -1)
        for row in range(self.rowCount()):
            for column in range(self.columnCount()):
                if not self.cellWidget(row, column):
                    empty_cell = (row, column)
                    break
            if empty_cell != (-1, -1):
                break
        if empty_cell == (-1, -1):
            self.insertRow(self.rowCount())
            empty_cell = (self.rowCount() - 1, 0)
        return empty_cell

    def count(self):
        """
        Returns the total number of characters that are already loaded on the table
        :return: int, total number of characters
        """

        return self.columnCount() * self.rowCount()

    def get_widgets(self):
        for row in range(self.rowCount()):
            for column in range(self.columnCount()):
                cell_widget = self.cellWidget(row, column)
                if not cell_widget:
                    continue
                item_widget = cell_widget.containedWidget
                yield item_widget



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widgets related with directories and files
"""

from __future__ import print_function, division, absolute_import

import os

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs import qt
from tpDcc.libs.python import path
from tpDcc.libs.qt.core import base, qtutils
from tpDcc.libs.qt.widgets import buttons


class FileListWidget(QListWidget, object):
    """
    Widgets that shows files and directories such Windows Explorer
    """

    directory_activated = Signal(str)
    file_activated = Signal(str)
    file_selected = Signal(str)
    folder_selected = Signal(str)
    directory_selected = Signal(str)
    files_selected = Signal(list)
    up_requested = Signal()
    update_requested = Signal()

    def __init__(self, parent):
        self.parent = parent
        super(FileListWidget, self).__init__(parent)

        # region Signals
        self.itemSelectionChanged.connect(self.selectItem)
        self.itemDoubleClicked.connect(self.activateItem)
        # endregion

    # region Override Functions
    def resizeEvent(self, event):
        """
        Overrides QWidget resizeEvent so when the widget is resize a update request signal is emitted
        :param event: QResizeEvent
        """

        self.update_requested.emit()
        super(FileListWidget, self).resizeEvent(event)

    def wheelEvent(self, event):
        """
        Overrides QWidget wheelEvent to smooth scroll bar movement
        :param event: QWheelEvent
        """

        sb = self.horizontalScrollBar()
        minValue = sb.minimum()
        maxValue = sb.maximum()
        if sb.isVisible() and maxValue > minValue:
            sb.setValue(sb.value() + (-1 if event.delta() > 0 else 1))
        super(FileListWidget, self).wheelEvent(event)

    def keyPressEvent(self, event):
        """
        Overrides QWidget keyPressEvent with some shortcuts when using the widget
        :param event:
        :return:
        """
        modifiers = event.modifiers()
        if event.key() == int(Qt.Key_Return) and modifiers == Qt.NoModifier:
            if len(self.selectedItems()) > 0:
                item = self.selectedItems()[0]
                if item.type() == 0:  # directory
                    self.directory_activated.emit(item.text())
                else:  # file
                    self.file_activated.emit(item.text())
        elif event.key() == int(Qt.Key_Backspace) and modifiers == Qt.NoModifier:
            self.up_requested.emit()
        elif event.key() == int(Qt.Key_F5) and modifiers == Qt.NoModifier:
            self.update_requested.emit()
        else:
            super(FileListWidget, self).keyPressEvent(event)
    # endregion

    # region Public Functions
    def selectItem(self):
        if len(self.selectedItems()) > 0:
            item = self.selectedItems()[0]
            if item.type() == 0:    # directory
                self.folder_selected.emit(item.text())
            if item.type() == 1:  # file
                self.file_selected.emit(item.text())

    def activateItem(self):
        if len(self.selectedItems()) > 0:
            item = self.selectedItems()[0]
            if item.type() == 0:  # directory
                self.directory_activated.emit(item.text())
            else:  # file
                self.file_activated.emit(item.text())

    def setExtendedSelection(self):
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.itemSelectionChanged.disconnect(self.selectItem)
        self.itemSelectionChanged.connect(self.processSelectionChanged)

    def processSelectionChanged(self):
        """
        Gets all selected items and emits a proper signal with the proper selected item names
        """

        items = filter(lambda x: x.type() != 0, self.selectedItems())
        names = map(lambda x: x.text(), items)
        self.files_selected.emit(names)
    # endregion


class FolderEditLine(QLineEdit, object):
    """
    Custom QLineEdit with drag and drop behaviour for files and folders
    """

    def __init__(self, parent=None):
        super(FolderEditLine, self).__init__(parent)

        self.setDragEnabled(True)
        self.setReadOnly(True)

    def dragEnterEvent(self, event):
        """
        Overrides QWidget dragEnterEvent to enable drop behaviour with file
        :param event: QDragEnterEvent
        :return:
        """
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dropEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            self.setText(urls[0].toLocalFile())


class SelectFolderButton(QWidget, object):
    """
    Button widget that allows to select folder paths
    """

    beforeNewDirectory = Signal()
    directoryChanged = Signal(object)   # Signal that is called when a new folder is selected

    def __init__(self, text='Browse', directory='', use_app_browser=False, parent=None):
        super(SelectFolderButton, self).__init__(parent)

        self._use_app_browser = use_app_browser
        self._directory = directory
        self.settings = None

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(2)
        self.setLayout(main_layout)

        folder_icon = tp.ResourcesMgr().icon('folder')
        self._folder_btn = buttons.IconButton(
            icon=folder_icon, icon_padding=2, button_style=buttons.ButtonStyles.FlatStyle)
        main_layout.addWidget(self._folder_btn)

        self._folder_btn.clicked.connect(self._open_folder_browser_dialog)

    # region Properties
    @property
    def folder_btn(self):
        return self._folder_btn

    def get_init_directory(self):
        return self._directory

    def set_init_directory(self, directory):
        self._directory = directory

    init_directory = property(get_init_directory, set_init_directory)
    # endregion

    # region Public Functions
    def set_settings(self, settings):
        self.settings = settings
    # endregion

    # region Private Functions
    def _open_folder_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        self.beforeNewDirectory.emit()

        if tp.Dcc.get_name() == tp.Dccs.Maya:
            import maya.cmds as cmds
            result = cmds.fileDialog2(caption='Select Folder', fileMode=3, startingDirectory=self.init_directory)
            if result:
                result = result[0]
            else:
                return
        else:
            raise NotImplementedError(
                'Open Folder Browser Dialog is not impelented in your current DCC: {}'.format(tp.Dcc.get_name()))

        self.directoryChanged.emit(result)
        # if not result or not os.path.isdir(result[0]):
        if not result or not os.path.isdir(result):
            return
        return path.clean_path(result[0])
    # endregion


class SelectFolder(QWidget, object):
    """
    Widget with button and line edit that opens a folder dialog to select folder paths
    """

    directoryChanged = Signal(object)  # Signal that is called when a new folder is selected

    def __init__(self, label_text='Select Folder', directory='', use_app_browser=False, use_icon=True, parent=None):
        super(SelectFolder, self).__init__(parent)

        self._use_app_browser = use_app_browser
        self._use_icon = use_icon
        self.settings = None
        self.directory = None
        self._label_text = label_text
        self._directory = directory

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(2)
        self.setLayout(main_layout)

        self._folder_label = QLabel(
            '{0}'.format(self._label_text)) if self._label_text == '' else QLabel('{0}:'.format(self._label_text))
        self._folder_line = FolderEditLine()
        self._folder_line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        if os.path.exists(self._directory):
            self._folder_line.setText(self._directory)

        if self._use_icon:
            folder_icon = tp.ResourcesMgr().icon('folder')
            self._folder_btn = buttons.IconButton(
                icon=folder_icon, icon_padding=2, button_style=buttons.ButtonStyles.FlatStyle)
        else:
            self._folder_btn = buttons.StyleBaseButton('Browse...')
        self._folder_btn.setMaximumHeight(20)

        for widget in [self._folder_label, self._folder_line, self._folder_btn]:
            main_layout.addWidget(widget)

        self._folder_btn.clicked.connect(self._open_folder_browser_dialog)

    @property
    def folder_label(self):
        return self._folder_label

    @property
    def folder_line(self):
        return self._folder_line

    @property
    def folder_btn(self):
        return self._folder_btn

    def set_directory_text(self, new_text):
        """
        Sets the text of the directory line
        :param new_text: str
        """

        self._folder_line.setText(new_text)

    def get_directory(self):
        """
        Returns directory set on the directory line
        :return: str
        """

        return str(self._folder_line.text())

    def set_directory(self, directory):
        """
        Sets the directory of the directory line
        """

        if not directory:
            return

        self.directory = directory

        self.set_directory_text(directory)

    def _open_folder_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        if tp.is_maya():
            import tpDcc.dccs.maya as maya
            result = maya.cmds.fileDialog2(
                caption='Select Folder', fileMode=3, startingDirectory=self.folder_line.text())
            if result:
                result = result[0]
            if not result or not os.path.isdir(result):
                return
            else:
                filename = path.clean_path(result)
                self.set_directory(filename)
                self._text_changed()
        else:
            raise NotImplementedError(
                'Open Folder Browser Dialog is not implemented in your current DCC: {}'.format(tp.Dcc.get_name()))

        return filename

    def _text_changed(self):
        """
        This function is called each time the user manually changes the line text
        Emits the signal to notify that the directory has changed
        :param directory: str, new edit line text after user edit
        """

        directory = self.get_directory()
        if path.is_dir(directory):
            self.directoryChanged.emit(directory)

    def _send_directories(self, directory):
        """
        Emit the directory changed signal with the given directory
        :param directory: str
        :return: str
        """

        self.directoryChanged.emit(directory)


class SelectFile(base.DirectoryWidget, object):
    """
    Widget with button and line edit that opens a file dialog to select file paths
    """

    directoryChanged = Signal(object)  # Signal that is called when a new folder is selected

    def __init__(self, label_text='Select File', directory='', use_app_browser=False,
                 filters=None, use_icon=True, parent=None):

        self._use_app_browser = use_app_browser
        self.settings = None
        self._use_icon = use_icon
        self._directory = directory
        self._label_text = label_text
        self._filters = filters

        super(SelectFile, self).__init__(parent)

    @property
    def file_label(self):
        return self._file_label

    @property
    def file_line(self):
        return self._file_line

    @property
    def file_btn(self):
        return self._folder_btn

    def get_main_layout(self):
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(2, 2, 2, 2)
        main_layout.setSpacing(2)

        return main_layout

    def ui(self):
        super(SelectFile, self).ui()

        self._file_label = QLabel('{0}'.format(self._label_text)) if self._label_text == '' else QLabel(
            '{0}:'.format(self._label_text))
        self._file_line = FolderEditLine()
        self._file_line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        if os.path.exists(self._directory):
            self._file_line.setText(self._directory)

        if self._use_icon:
            folder_icon = tp.ResourcesMgr().icon('folder')
            self._file_btn = buttons.IconButton(
                icon=folder_icon, icon_padding=2, button_style=buttons.ButtonStyles.FlatStyle)
        else:
            self._file_btn = buttons.StyleBaseButton('Browse ...')

        for widget in [self._file_label, self._file_line, self._file_btn]:
            self.main_layout.addWidget(widget)

    def setup_signals(self):
        self._file_btn.clicked.connect(self._open_file_browser_dialog)
        self._file_line.textChanged.connect(self._text_changed)

    def set_settings(self, settings):
        """
        Set new settings. Override in new classes to add custom behaviour
        :param settings:
        :return:
        """
        self.settings = settings

    def update_settings(self, filename):
        """
        Updates current settings. Override in new classes to add custom behaviour
        :param settings: new selected path for the user
        """

        pass

    def set_label(self, text):
        """
        Sets the directory label text
        :param text: str, new directory label text
        :return:
        """

        self._file_label.setText(text)

    def set_directory(self, directory):
        """
        Sets the text of the directory line
        :param directory: str
        """

        super(SelectFile, self).set_directory(directory=directory)

        self._file_line.setText(directory)

    def get_directory(self):
        """
        Returns directory setted on the directory line
        :return: str
        """

        return self._file_line.text()

    def _open_file_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        if tp.is_maya():
            import tpDcc.dccs.maya as maya

            file_line = self._file_line.text()
            if os.path.isfile(file_line):
                file_line = os.path.dirname(file_line)
            result = maya.cmds.fileDialog2(
                caption='Select File', fileMode=1, fileFilter=self._filters or '', startingDirectory=file_line)
            if result:
                result = result[0]
            if not result or not os.path.isfile(result):
                qt.logger.warning('Selected file {} is not a valid file!'.format(result))
                return
            else:
                filename = path.clean_path(result)
                self._file_line.setText(filename)
                self.directoryChanged.emit(filename)
                self.update_settings(filename=filename)
        else:
            raise NotImplementedError(
                'Select File Dialog is not implemented in your current DCC: {}'.format(tp.Dcc.get_name()))

        return filename

    def _text_changed(self):
        """
        This function is called each time the user manually changes the line text
        :param directory: str, new edit line text after user edit
        """

        f = self.get_directory()
        if path.is_file(f):
            self.directoryChanged.emit(f)


class GetDirectoryWidget(base.DirectoryWidget, object):
    directoryChanged = Signal(object)
    textChanged = Signal(object)

    def __init__(self, parent=None):

        self._label = 'directory'
        self._show_files = False

        super(GetDirectoryWidget, self).__init__(parent=parent)

    @property
    def show_files(self):
        return self._show_files

    def ui(self):
        super(GetDirectoryWidget, self).ui()

        self._directory_widget = QWidget()
        directory_layout = QHBoxLayout()
        self._directory_widget.setLayout(directory_layout)
        self.main_layout.addWidget(self._directory_widget)

        self._directory_lbl = QLabel('directory')
        self._directory_lbl.setMinimumWidth(60)
        self._directory_lbl.setMaximumWidth(100)
        self._directory_edit = QLineEdit()
        self._directory_edit.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self._directory_browse_btn = QPushButton('browse')

        directory_layout.addWidget(self._directory_lbl)
        directory_layout.addWidget(self._directory_edit)
        directory_layout.addWidget(self._directory_browse_btn)

    def setup_signals(self):
        self._directory_edit.textChanged.connect(self._on_text_changed)
        self._directory_browse_btn.clicked.connect(self._on_browse)

    def set_directory(self, directory):
        super(GetDirectoryWidget, self).set_directory(directory)
        self.set_directory_text(directory)

    def get_directory(self):
        return self._directory_edit.text()

    def set_label(self, label):
        length = len(label) * 8
        self._directory_lbl.setMinimumWidth(length)
        self._directory_lbl.setText(label)

    def set_directory_text(self, text):
        self._directory_edit.setText(text)

    def set_placeholder_text(self, text):
        self._directory_edit.setPlaceholderText(text)

    def set_example(self, text):
        self.set_placeholder_text('example: {}'.format(text))

    def set_error(self, flag):

        if tp.is_maya():
            yes_color = QColor(0, 255, 0, 50)
            no_color = QColor(255, 0, 0, 50)
        else:
            yes_color = QColor(200, 255, 200, 100)
            no_color = QColor(25, 200, 200, 100)

        palette = QPalette()
        if flag:
            palette.setColor(QPalette().Base, no_color)
        else:
            palette.setColor(QPalette().Base, yes_color)

        self._directory_edit.setPalette(palette)

    def _on_text_changed(self, text):
        directory = self.get_directory()
        if os.path.exists(directory):
            self.set_error(False)
        else:
            self.set_error(True)

        if not text:
            self._directory_edit.setPalette(QLineEdit().palette())

        self.directoryChanged.emit(directory)

    def _on_browse(self):
        directory = self.get_directory()
        if not directory:
            placeholder = self._directory_edit.placeholderText()
            if placeholder and placeholder.startswith('example: '):
                example_path = placeholder[0]
                if os.path.exists(example_path):
                    directory = example_path

        filename = qtutils.get_folder(directory, show_files=self._show_files, parent=self)
        filename = path.clean_path(filename)
        if filename and path.is_dir(filename):
            self._directory_edit.setText(filename)
            self.directoryChanged.emit(filename)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base classes for list views
"""

from __future__ import print_function, division, absolute_import

import os
import glob

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs.python import python
from tpDcc.libs.qt.widgets import models


class BaseListView(QListView, object):
    """
    Base class for list views
    """

    def __ini__(self, parent=None):
        super(BaseListView, self).__init__(parent=parent)


class ListView(BaseListView, object):
    """
    Extends basic QListView functionality
    """

    current_selection_changed = Signal(object, object)
    middle_clicked = Signal(QModelIndex)

    def __init__(self, parent=None):
        super(ListView, self).__init__(parent)

    # region Public Functions
    def appendItem(self, item):
        """
        Adds a new item to the list view model
        :param item: variant
        """

        if self.model():
            self.model().appendItem(item)
    # endregion

    # region Override Functions
    def selectionChanged(self, selected, deselected):
        """
        Override selectionChanged behavior
        :param selected: list<variant>, new selection
        :param deselected: list<variant>, previous selection
        :warning: This method is override to avoid a Qt 4.7.0 QModelIndexList destructor crash.
            See ItemSelection for more information.
        """

        self.current_selection_changed.emit(models.ItemSelection(selected), models.ItemSelection(deselected))
        super(ListView, self).selectionChanged(models.ItemSelection(selected), models.ItemSelection(deselected))

    def selectedIndexes(self):
        """
        Override selectedIndexes behavior
        :warning: This method is override to avoid  a Qt 4.7.0 QModelIndexList destructor crash.
            See ItemSelection for more information.
        """

        item_selection = models.ItemSelection(self.selectionModel().selection())
        return item_selection.indexes()

    def mousePressEvent(self, event):
        """
        Override mousePressEvent behavior to add a custom middle click signal
        :param event: QMouseEvent
        """

        if event.button() == Qt.MidButton:
            model_index = self.indexAt(event.pos())
            if model_index.row() >= 0:
                self.middle_clicked.emit(model_index)
        super(ListView, self).mousePressEvent(event)
    # endregion


class DraggableListView(ListView, object):
    """
    Extends ListView functionality width draggable functionality
    """

    addedItemDir = Signal(list)

    def __init__(self, parent=None):
        super(DraggableListView, self).__init__(parent=parent)

        self._accepted_files = list()

        self.setEditTriggers(QListView.NoEditTriggers)
        self.setDragEnabled(True)
        self.setDragDropMode(QListView.DragOnly)
        self.setDefaultDropAction(Qt.IgnoreAction)
        self.setAcceptDrops(True)
        self.setIconSize(QSize(120, 120))
        self.setViewMode(QListView.IconMode)
        self.setMovement(QListView.Snap)
        self.setResizeMode(QListView.Adjust)
        self.setUniformItemSizes(True)
        self.setMouseTracking(True)

    def startDrag(self, supported_actions):
        """
        Function that is called when we start dragging an item
        This function set ups the mime data and creates a copy of the image
        of the item that is show in cursor position while dragging action
        is enabled
        """

        index = self.currentIndex()
        model = self.model()
        item = model.itemFromIndex(index)
        drag = QDrag(self)
        mime_data = QMimeData()
        mime_data.setData('item_text', QByteArray(str(item.text())))
        mime_data.setData('item_size', QByteArray('%.1f %.1f' % (item.size.width(), item.size.height())))
        url = QUrl.fromLocalFile(item.text())
        mime_data.setUrls(url)
        drag.setMimeData(mime_data)
        pixmap = item.icon().pixmap(50, 50)
        drag.setDragCursor(pixmap, Qt.CopyAction)
        drag.start()

    def dragEnterEvent(self, event):
        mime_data = event.mimeData()
        if mime_data.hasUrls():
            for path in map(lambda x: x.toLocalFile(), mime_data.urls()):
                if os.path.isdir(path) or os.path.isfile(path) and \
                        os.path.splitext(path)[-1].lower() in self._accepted_files:
                    self.setDragDropMode(QListView.InternalMove)
                    event.accept()
                    break
        super(DraggableListView, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        event.accept()
        super(DraggableListView, self).dragMoveEvent(event)

    def dropEvent(self, event):
        mime_data = event.mimeData()
        if mime_data.hasUrls():
            send_data = list()
            for path in map(lambda x: x.toLocalFile(), mime_data.urls()):
                for file in self._accepted_files:
                    if os.path.isdir(path) and glob.glob(os.path.join(path, '*' + file))\
                            or os.path.isfile(path) and os.path.splitext(path)[-1].lower() == file:
                        send_data.append(path)
            if send_data:
                self.addedItemDir.emit(send_data)
            self.setDragEnabled(True)
            self.setDragDropMode(QListView.DragOnly)
            self.setDefaultDropAction(Qt.IgnoreAction)
            self.setAcceptDrops(True)
            super(DraggableListView, self).dropEvent(event)

    def set_accepted_files(self, list_files):

        list_files = python.force_list(list_files)

        for f in list_files:
            if not f.startswith('.'):
                f = '.' + f
            self._accepted_files.append(f)


class BaseTableView(QTableView, object):
    def __init__(self, parent=None, **kwargs):
        super(BaseTableView, self).__init__(parent=parent, **kwargs)

        self._last_indexes = list()

        self.installEventFilter(self)

    def focusOutEvent(self, event):
        if self.selectionModel().selectedIndexes():
            for index in self.selectionModel().selectedRows():
                self._last_indexes.append(QPersistentModelIndex(index))
        if self._last_indexes:
            for i in self._last_indexes:
                self.selectionModel().setCurrentIndex(i, QItemSelectionModel.Select)
        event.accept()

    def get_selected_indexes(self):
        """
        Returns the selected indexes
        """

        return self.selectionModel().selectedIndexes()

    def get_selected_rows(self):
        """
        Returns the selected rows
        """

        return self.selectionModel().selectedRows()


class BaseTreeView(QTreeView, object):
    """
    Base class for tree views
    """

    def __init__(self, parent=None):
        super(BaseTreeView, self).__init__(parent=parent)

    def expand_tree(self, root_node):
        """
        Expands all the collapsed elements in a tree starting at the rootNode
        :param root_node: Start node from which we start expanding the tree
        """

        parent = root_node.parent()
        parent_id = self.model().createIndex(parent.row(), 0, parent) if parent else QModelIndex()
        index = self.model().index(root_node.row(), 0, parent_id)
        self.setExpanded(index, True)
        for child in root_node.children:
            self.expand_tree(child)
